# Low-level Hardware Info. Expressed using NetSquid concepts and objects.
from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Type

from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
    IMeasure,
)
from netsquid.components.instructions import Instruction as NetSquidInstruction
from netsquid.components.models.qerrormodels import (
    DepolarNoiseModel,
    QuantumErrorModel,
    T1T2NoiseModel,
)
from netsquid_magic.state_delivery_sampler import (
    DepolariseWithFailureStateSamplerFactory,
    IStateDeliverySamplerFactory,
    PerfectStateSamplerFactory,
)

from qoala.lang.common import MultiQubit

# A measurement that is meant to take 0 time.
# Used in e.g. Measure Directly EPR generation.
INSTR_MEASURE_INSTANT = IMeasure("measurement_instant_op")

# Config Interface


class LhiQubitConfigInterface(ABC):
    @abstractmethod
    def to_is_communication(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def to_error_model(self) -> Type[QuantumErrorModel]:
        raise NotImplementedError

    @abstractmethod
    def to_error_model_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError


class LhiGateConfigInterface(ABC):
    @abstractmethod
    def to_instruction(self) -> Type[NetSquidInstruction]:
        raise NotImplementedError

    @abstractmethod
    def to_duration(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def to_error_model(self) -> Type[QuantumErrorModel]:
        raise NotImplementedError

    @abstractmethod
    def to_error_model_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError


class LhiTopologyConfigInterface(ABC):
    @abstractmethod
    def get_qubit_configs(self) -> Dict[int, LhiQubitConfigInterface]:
        raise NotImplementedError

    @abstractmethod
    def get_single_gate_configs(self) -> Dict[int, List[LhiGateConfigInterface]]:
        raise NotImplementedError

    @abstractmethod
    def get_multi_gate_configs(
        self,
    ) -> Dict[MultiQubit, List[LhiGateConfigInterface]]:
        raise NotImplementedError


# Data classes


@dataclass
class LhiQubitInfo:
    is_communication: bool
    error_model: Type[QuantumErrorModel]
    error_model_kwargs: Dict[str, Any]


@dataclass
class LhiGateInfo:
    instruction: Type[NetSquidInstruction]
    duration: float  # ns
    error_model: Type[QuantumErrorModel]
    error_model_kwargs: Dict[str, Any]


@dataclass
class LhiTopology:
    qubit_infos: Dict[int, LhiQubitInfo]  # qubit ID -> info
    single_gate_infos: Dict[int, List[LhiGateInfo]]  # qubit ID -> gates
    multi_gate_infos: Dict[
        MultiQubit, List[LhiGateInfo]
    ]  # ordered qubit ID list -> gates

    def find_single_gate(
        self, qubit_id: int, instr: Type[NetSquidInstruction]
    ) -> Optional[LhiGateInfo]:
        if qubit_id not in self.single_gate_infos:
            return None
        for info in self.single_gate_infos[qubit_id]:
            if info.instruction == instr:
                return info
        return None

    def find_multi_gate(
        self, qubit_ids: List[int], instr: Type[NetSquidInstruction]
    ) -> Optional[LhiGateInfo]:
        multi = MultiQubit(qubit_ids)
        if multi not in self.multi_gate_infos:
            return None
        for info in self.multi_gate_infos[multi]:
            if info.instruction == instr:
                return info
        return None


# Convenience methods.


class LhiTopologyBuilder:
    """Convenience methods for creating a Topology object."""

    @classmethod
    def from_config(cls, cfg: LhiTopologyConfigInterface) -> LhiTopology:
        qubit_infos: Dict[int, LhiQubitInfo] = {}
        for i, cfg_info in cfg.get_qubit_configs().items():
            qubit_infos[i] = LhiQubitInfo(
                is_communication=cfg_info.to_is_communication(),
                error_model=cfg_info.to_error_model(),
                error_model_kwargs=cfg_info.to_error_model_kwargs(),
            )

        single_gate_infos: Dict[int, List[LhiGateInfo]] = {}
        for i, cfg_infos in cfg.get_single_gate_configs().items():
            single_gate_infos[i] = [
                LhiGateInfo(
                    instruction=info.to_instruction(),
                    duration=info.to_duration(),
                    error_model=info.to_error_model(),
                    error_model_kwargs=info.to_error_model_kwargs(),
                )
                for info in cfg_infos
            ]

        multi_gate_infos: Dict[MultiQubit, List[LhiGateInfo]] = {}
        for ids, cfg_infos in cfg.get_multi_gate_configs().items():
            multi_gate_infos[ids] = [
                LhiGateInfo(
                    instruction=info.to_instruction(),
                    duration=info.to_duration(),
                    error_model=info.to_error_model(),
                    error_model_kwargs=info.to_error_model_kwargs(),
                )
                for info in cfg_infos
            ]

        return LhiTopology(
            qubit_infos=qubit_infos,
            single_gate_infos=single_gate_infos,
            multi_gate_infos=multi_gate_infos,
        )

    @classmethod
    def t1t2_qubit(cls, is_communication: bool, t1: float, t2: float) -> LhiQubitInfo:
        return LhiQubitInfo(
            is_communication=is_communication,
            error_model=T1T2NoiseModel,
            error_model_kwargs={"T1": t1, "T2": t2},
        )

    @classmethod
    def perfect_qubit(cls, is_communication: bool) -> LhiQubitInfo:
        return cls.t1t2_qubit(is_communication=is_communication, t1=0, t2=0)

    @classmethod
    def depolar_gates(
        cls,
        duration: float,
        instructions: List[NetSquidInstruction],
        depolar_rate: float,
    ) -> List[LhiGateInfo]:
        return [
            LhiGateInfo(
                instruction=instr,
                duration=duration,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": depolar_rate},
            )
            for instr in instructions
        ]

    @classmethod
    def perfect_gates(
        cls, duration: float, instructions: List[NetSquidInstruction]
    ) -> List[LhiGateInfo]:
        return cls.depolar_gates(
            duration=duration, instructions=instructions, depolar_rate=0
        )

    @classmethod
    def perfect_uniform_default_gates(cls, num_qubits) -> LhiTopology:
        # TODO: test this and update default values
        return cls.perfect_uniform(
            num_qubits=num_qubits,
            single_instructions=[
                INSTR_INIT,
                INSTR_X,
                INSTR_Y,
                INSTR_Z,
                INSTR_H,
                INSTR_ROT_X,
                INSTR_ROT_Y,
                INSTR_ROT_Z,
                INSTR_MEASURE,
                INSTR_MEASURE_INSTANT,
            ],
            single_duration=5e3,
            two_instructions=[INSTR_CNOT, INSTR_CZ],
            two_duration=100e3,
        )

    @classmethod
    def perfect_uniform(
        cls,
        num_qubits,
        single_instructions: List[NetSquidInstruction],
        single_duration: float,
        two_instructions: List[NetSquidInstruction],
        two_duration: float,
    ) -> LhiTopology:
        return cls.fully_uniform(
            num_qubits=num_qubits,
            qubit_info=cls.perfect_qubit(is_communication=True),
            single_gate_infos=cls.perfect_gates(single_duration, single_instructions),
            two_gate_infos=cls.perfect_gates(two_duration, two_instructions),
        )

    @classmethod
    def fully_uniform(
        cls,
        num_qubits,
        qubit_info: LhiQubitInfo,
        single_gate_infos: List[LhiGateInfo],
        two_gate_infos: List[LhiGateInfo],
    ) -> LhiTopology:
        q_infos = {i: qubit_info for i in range(num_qubits)}
        sg_infos = {i: single_gate_infos for i in range(num_qubits)}
        mg_infos = {}
        for i in range(num_qubits):
            for j in range(num_qubits):
                if i != j:
                    multi = MultiQubit([i, j])
                    mg_infos[multi] = two_gate_infos
        return LhiTopology(q_infos, sg_infos, mg_infos)

    @classmethod
    def perfect_star(
        cls,
        num_qubits: int,
        comm_instructions: List[NetSquidInstruction],
        comm_duration: float,
        mem_instructions: List[NetSquidInstruction],
        mem_duration: float,
        two_instructions: List[NetSquidInstruction],
        two_duration: float,
    ) -> LhiTopology:
        comm_qubit_info = cls.perfect_qubit(is_communication=True)
        mem_qubit_info = cls.perfect_qubit(is_communication=False)
        comm_gate_infos = cls.perfect_gates(comm_duration, comm_instructions)
        mem_gate_infos = cls.perfect_gates(mem_duration, mem_instructions)
        two_gate_infos = cls.perfect_gates(two_duration, two_instructions)

        q_infos = {0: comm_qubit_info}
        for i in range(1, num_qubits):
            q_infos[i] = mem_qubit_info

        sg_infos = {0: comm_gate_infos}
        for i in range(1, num_qubits):
            sg_infos[i] = mem_gate_infos

        mg_infos = {}
        for i in range(1, num_qubits):
            mg_infos[MultiQubit([0, i])] = two_gate_infos

        return LhiTopology(q_infos, sg_infos, mg_infos)

    @classmethod
    def generic_t1t2_star(
        cls,
        num_qubits: int,
        comm_t1: float,
        comm_t2: float,
        mem_t1: float,
        mem_t2: float,
        comm_instructions: List[NetSquidInstruction],
        comm_duration: float,
        comm_instr_depolar_rate: float,
        mem_instructions: List[NetSquidInstruction],
        mem_duration: float,
        mem_instr_depolar_rate: float,
        two_instructions: List[NetSquidInstruction],
        two_duration: float,
        two_instr_depolar_rate: float,
    ) -> LhiTopology:
        comm_qubit_info = cls.t1t2_qubit(is_communication=True, t1=comm_t1, t2=comm_t2)
        mem_qubit_info = cls.t1t2_qubit(is_communication=False, t1=mem_t1, t2=mem_t2)

        comm_gate_infos = cls.depolar_gates(
            comm_duration, comm_instructions, comm_instr_depolar_rate
        )
        mem_gate_infos = cls.depolar_gates(
            mem_duration, mem_instructions, mem_instr_depolar_rate
        )
        two_gate_infos = cls.depolar_gates(
            two_duration, two_instructions, two_instr_depolar_rate
        )

        q_infos = {0: comm_qubit_info}
        for i in range(1, num_qubits):
            q_infos[i] = mem_qubit_info

        sg_infos = {0: comm_gate_infos}
        for i in range(1, num_qubits):
            sg_infos[i] = mem_gate_infos

        mg_infos = {}
        for i in range(1, num_qubits):
            mg_infos[MultiQubit([0, i])] = two_gate_infos

        return LhiTopology(q_infos, sg_infos, mg_infos)


class LhiLatenciesConfigInterface(ABC):
    @abstractmethod
    def get_host_instr_time(self) -> float:
        raise NotImplementedError

    @abstractmethod
    def get_qnos_instr_time(self) -> float:
        raise NotImplementedError

    @abstractmethod
    def get_host_peer_latency(self) -> float:
        raise NotImplementedError


@dataclass
class LhiLatencies:
    host_instr_time: float = 0  # duration of classical Host instr execution
    qnos_instr_time: float = 0  # duration of classical Qnos instr execution
    host_peer_latency: float = 0  # processing time for Host messages from remote node

    @classmethod
    def from_config(cls, cfg: LhiLatenciesConfigInterface) -> LhiLatencies:
        return LhiLatencies(
            host_instr_time=cfg.get_host_instr_time(),
            qnos_instr_time=cfg.get_qnos_instr_time(),
            host_peer_latency=cfg.get_host_peer_latency(),
        )

    @classmethod
    def all_zero(cls) -> LhiLatencies:
        # NOTE: can also just use LhiLatencies() which will default all values to 0
        # However, using this classmethod makes this behavior more explicit and clear.
        return LhiLatencies(0, 0, 0)


class LhiLinkConfigInterface(ABC):
    @abstractmethod
    def to_sampler_factory(self) -> Type[IStateDeliverySamplerFactory]:
        raise NotImplementedError

    @abstractmethod
    def to_sampler_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def to_state_delay(self) -> float:
        raise NotImplementedError


@dataclass
class LhiLinkInfo:
    sampler_factory: Type[IStateDeliverySamplerFactory]
    sampler_kwargs: Dict[str, Any]
    state_delay: float  # time between EPR generation and putting the state into memory

    @classmethod
    def from_config(cls, cfg: LhiLinkConfigInterface) -> LhiLinkInfo:
        return LhiLinkInfo(
            sampler_factory=cfg.to_sampler_factory(),
            sampler_kwargs=cfg.to_sampler_kwargs(),
            state_delay=cfg.to_state_delay(),
        )

    @classmethod
    def perfect(cls, duration: float) -> LhiLinkInfo:
        return LhiLinkInfo(
            sampler_factory=PerfectStateSamplerFactory,
            sampler_kwargs={"cycle_time": 0},
            state_delay=duration,
        )

    @classmethod
    def depolarise(
        cls,
        cycle_time: float,
        prob_max_mixed: float,
        prob_success: float,
        state_delay: float,
    ) -> LhiLinkInfo:
        return LhiLinkInfo(
            sampler_factory=DepolariseWithFailureStateSamplerFactory,
            sampler_kwargs={
                "cycle_time": cycle_time,
                "prob_max_mixed": prob_max_mixed,
                "prob_success": prob_success,
            },
            state_delay=state_delay,
        )


@dataclass
class LhiNetworkInfo:
    # (node A ID, node B ID) -> link info
    # for a pair (a, b) there exists no separate (b, a) info (it is the same)
    links: Dict[Tuple[int, int], LhiLinkInfo]

    @classmethod
    def fully_connected(cls, node_ids: List[int], info: LhiLinkInfo) -> LhiNetworkInfo:
        links: Dict[Tuple[int, int], LhiLinkInfo] = {}
        for n1, n2 in itertools.combinations(node_ids, 2):
            links[(n1, n2)] = info
        return LhiNetworkInfo(links)

    @classmethod
    def perfect_fully_connected(
        cls, node_ids: List[int], duration: float
    ) -> LhiNetworkInfo:
        link = LhiLinkInfo.perfect(duration)
        return cls.fully_connected(node_ids, link)

    def get_link(self, node_id1: int, node_id2: int) -> LhiLinkInfo:
        try:
            return self.links[(node_id1, node_id2)]
        except KeyError:
            return self.links[(node_id2, node_id1)]


@dataclass
class LhiProcNodeInfo:
    id: int
    name: str
    topology: LhiTopology
    latencies: LhiLatencies



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional

import netsquid as ns
from netqasm.lang.parsing import parse_text_subroutine
from netqasm.lang.subroutine import Subroutine

from pydynaa import EventExpression
from qoala.lang.ehi import EhiBuilder, UnitModule
from qoala.lang.hostlang import (
    AddCValueOp,
    AssignCValueOp,
    BasicBlock,
    BasicBlockType,
    BitConditionalMultiplyConstantCValueOp,
    ClassicalIqoalaOp,
    IqoalaVector,
    MultiplyConstantCValueOp,
    ReceiveCMsgOp,
    ReturnResultOp,
    RunRequestOp,
    RunSubroutineOp,
    SendCMsgOp,
)
from qoala.lang.program import ProgramMeta, QoalaProgram
from qoala.lang.request import (
    CallbackType,
    EprRole,
    EprType,
    QoalaRequest,
    RequestRoutine,
    RequestVirtIdMapping,
)
from qoala.lang.routine import LocalRoutine, RoutineMetadata
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.message import LrCallTuple, Message, RrCallTuple
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.sharedmem import SharedMemory
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.hostinterface import HostInterface, HostLatencies
from qoala.sim.host.hostprocessor import HostProcessor
from qoala.sim.process import QoalaProcess
from qoala.util.tests import netsquid_run, yield_from

MOCK_MESSAGE = Message(content=42)
MOCK_QNOS_RET_REG = "R0"
MOCK_QNOS_RET_VALUE = 7
MOCK_NETSTACK_RET_VALUE = 22


@dataclass(eq=True, frozen=True)
class InterfaceEvent:
    peer: str
    msg: Message


class MockHostInterface(HostInterface):
    def __init__(self, shared_mem: Optional[SharedMemory] = None) -> None:
        self.send_events: List[InterfaceEvent] = []
        self.recv_events: List[InterfaceEvent] = []

        self.shared_mem = shared_mem

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self.send_events.append(InterfaceEvent(peer, msg))

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent(peer, MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_qnos_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("qnos", msg))

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("qnos", MOCK_MESSAGE))
        if self.shared_mem is not None:
            # Hack to find out which addr was allocated to write results to.
            result_addr = self.shared_mem._lr_out_addrs[0]
            self.shared_mem.write_lr_out(result_addr, [MOCK_QNOS_RET_VALUE])
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_netstack_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("netstack", msg))

    def receive_netstack_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("netstack", MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    @property
    def name(self) -> str:
        return "mock"


def create_program(
    instrs: Optional[List[ClassicalIqoalaOp]] = None,
    subroutines: Optional[Dict[str, LocalRoutine]] = None,
    requests: Optional[Dict[str, RequestRoutine]] = None,
    meta: Optional[ProgramMeta] = None,
) -> QoalaProgram:
    if instrs is None:
        instrs = []
    if subroutines is None:
        subroutines = {}
    if requests is None:
        requests = {}
    if meta is None:
        meta = ProgramMeta.empty("prog")
    # TODO: split into proper blocks
    block = BasicBlock("b0", BasicBlockType.CL, instrs)
    return QoalaProgram(
        blocks=[block], local_routines=subroutines, request_routines=requests, meta=meta
    )


def create_process(
    program: QoalaProgram,
    interface: HostInterface,
    inputs: Optional[Dict[str, Any]] = None,
) -> QoalaProcess:
    if inputs is None:
        inputs = {}
    prog_input = ProgramInput(values=inputs)

    mock_ehi = EhiBuilder.perfect_uniform(
        num_qubits=2,
        flavour=None,
        single_instructions=[],
        single_duration=0,
        two_instructions=[],
        two_duration=0,
    )

    instance = ProgramInstance(
        pid=0,
        program=program,
        inputs=prog_input,
        unit_module=UnitModule.from_full_ehi(mock_ehi),
        block_tasks=[],
    )

    mem = ProgramMemory(pid=0)
    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={
            id: ClassicalSocket(interface, name)
            for (id, name) in program.meta.csockets.items()
        },
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def test_initialize():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program()
    process = create_process(
        program, interface, inputs={"x": 1, "theta": 3.14, "name": "alice"}
    )

    processor.initialize(process)
    host_mem = process.prog_memory.host_mem
    assert host_mem.read("x") == 1
    assert host_mem.read("theta") == 3.14
    assert host_mem.read("name") == "alice"


def test_assign_cvalue():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(instrs=[AssignCValueOp("x", 3)])
    process = create_process(program, interface)
    processor.initialize(process)

    yield_from(processor.assign_instr_index(process, 0))
    assert process.prog_memory.host_mem.read("x") == 3


def test_assign_cvalue_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    latencies = HostLatencies(host_instr_time=500)
    processor = HostProcessor(interface, latencies)
    program = create_program(instrs=[AssignCValueOp("x", 3)])
    process = create_process(program, interface)
    processor.initialize(process)

    assert ns.sim_time() == 0
    netsquid_run(processor.assign_instr_index(process, 0))
    assert process.prog_memory.host_mem.read("x") == 3
    assert ns.sim_time() == 500


def test_send_msg():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}
    program = create_program(instrs=[SendCMsgOp("bob", "msg")], meta=meta)
    process = create_process(program, interface, inputs={"bob": 0, "msg": 12})
    processor.initialize(process)

    yield_from(processor.assign_instr_index(process, 0))
    assert interface.send_events[0] == InterfaceEvent("bob", Message(content=12))


def test_recv_msg():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}
    program = create_program(instrs=[ReceiveCMsgOp("bob", "msg")], meta=meta)
    process = create_process(program, interface, inputs={"bob": 0})
    processor.initialize(process)

    yield_from(processor.assign_instr_index(process, 0))
    assert interface.recv_events[0] == InterfaceEvent("bob", MOCK_MESSAGE)
    assert process.prog_memory.host_mem.read("msg") == MOCK_MESSAGE.content


def test_recv_msg_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    latencies = HostLatencies(host_instr_time=500, host_peer_latency=1e6)
    processor = HostProcessor(interface, latencies)
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}
    program = create_program(instrs=[ReceiveCMsgOp("bob", "msg")], meta=meta)
    process = create_process(program, interface, inputs={"bob": 0})
    processor.initialize(process)

    assert ns.sim_time() == 0
    netsquid_run(processor.assign_instr_index(process, 0))
    assert interface.recv_events[0] == InterfaceEvent("bob", MOCK_MESSAGE)
    assert process.prog_memory.host_mem.read("msg") == MOCK_MESSAGE.content
    assert ns.sim_time() == 1e6  # no host_instr_time used !


def test_add_cvalue():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(
        instrs=[
            AssignCValueOp("a", 2),
            AssignCValueOp("b", 3),
            AddCValueOp("sum", "a", "b"),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("sum") == 5


def test_add_cvalue_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies(host_instr_time=1200))
    program = create_program(
        instrs=[
            AssignCValueOp("a", 2),
            AssignCValueOp("b", 3),
            AddCValueOp("sum", "a", "b"),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    assert ns.sim_time() == 0
    for i in range(len(program.instructions)):
        netsquid_run(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("sum") == 5
    assert ns.sim_time() == len(program.instructions) * 1200


def test_multiply_const():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(
        instrs=[AssignCValueOp("a", 4), MultiplyConstantCValueOp("result", "a", -1)]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result") == -4


def test_bit_cond_mult():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(
        instrs=[
            AssignCValueOp("var1", 4),
            AssignCValueOp("var2", 7),
            AssignCValueOp("cond1", 0),
            AssignCValueOp("cond2", 1),
            BitConditionalMultiplyConstantCValueOp("result1", "var1", "cond1", -1),
            BitConditionalMultiplyConstantCValueOp("result2", "var2", "cond2", -1),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result1") == 4
    assert process.prog_memory.host_mem.read("result2") == -7


def test_run_subroutine():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero(), asynchronous=True)

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    routine = LocalRoutine("subrt1", subrt, return_vars=[], metadata=metadata)

    program = create_program(
        instrs=[RunSubroutineOp(None, IqoalaVector([]), "subrt1")],
        subroutines={"subrt1": routine},
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    # Host processor should have communicated with the qnos processor.
    send_event = interface.send_events[0]
    assert send_event.peer == "qnos"
    assert isinstance(send_event.msg, Message)
    assert isinstance(send_event.msg.content, LrCallTuple)
    assert send_event.msg.content.routine_name == "subrt1"
    assert interface.recv_events[0] == InterfaceEvent("qnos", MOCK_MESSAGE)


def test_run_subroutine_2():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero(), asynchronous=True)

    subrt_text = """
    set R0 {my_value}
    ret_reg R0
    """
    subrt = parse_text_subroutine(subrt_text)
    routine = LocalRoutine(
        "subrt1",
        subrt,
        return_vars=["m"],
        metadata=RoutineMetadata.use_none(),
    )

    program = create_program(
        instrs=[
            AssignCValueOp("my_value", 16),
            RunSubroutineOp(IqoalaVector(["m"]), IqoalaVector(["my_value"]), "subrt1"),
        ],
        subroutines={"subrt1": routine},
    )
    process = create_process(program, interface)

    # Make sure interface can mimick writing subroutine results to shared memory
    interface.shared_mem = process.prog_memory.shared_mem

    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    send_event = interface.send_events[0]
    assert send_event.peer == "qnos"
    assert isinstance(send_event.msg, Message)
    assert isinstance(send_event.msg.content, LrCallTuple)
    assert send_event.msg.content.routine_name == "subrt1"
    assert interface.recv_events[0] == InterfaceEvent("qnos", MOCK_MESSAGE)

    # Hack to find out which addr was allocated to write results to.
    result_addr = process.prog_memory.shared_mem._lr_out_addrs[0]
    assert process.prog_memory.shared_mem.read_lr_out(result_addr, 1) == [
        MOCK_QNOS_RET_VALUE
    ]

    # Check if result was correctly written to host variable "m".
    assert process.prog_memory.host_mem.read("m") == MOCK_QNOS_RET_VALUE


def create_simple_request(
    remote_id: int,
    num_pairs: int,
    virt_ids: RequestVirtIdMapping,
    typ: EprType,
    role: EprRole,
) -> QoalaRequest:
    return QoalaRequest(
        name="req",
        remote_id=remote_id,
        epr_socket_id=0,
        num_pairs=num_pairs,
        virt_ids=virt_ids,
        timeout=1000,
        fidelity=0.65,
        typ=typ,
        role=role,
    )


def test_run_request():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero(), asynchronous=True)

    request = create_simple_request(
        remote_id=2,
        num_pairs=5,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.CREATE_KEEP,
        role=EprRole.CREATE,
    )
    routine = RequestRoutine("req", request, [], CallbackType.WAIT_ALL, None)

    program = create_program(
        instrs=[RunRequestOp(None, IqoalaVector([]), "req")],
        requests={"req": routine},
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    # Host processor should have communicated with the netstack processor.
    send_event = interface.send_events[0]
    assert send_event.peer == "netstack"
    assert isinstance(send_event.msg, Message)
    assert isinstance(send_event.msg.content, RrCallTuple)
    assert send_event.msg.content.routine_name == "req"
    assert interface.recv_events[0] == InterfaceEvent("netstack", MOCK_MESSAGE)


def test_return_result():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(
        instrs=[AssignCValueOp("result", 2), ReturnResultOp("result")]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result") == 2
    assert process.result.values == {"result": 2}


if __name__ == "__main__":
    test_initialize()
    test_assign_cvalue()
    test_assign_cvalue_with_latencies()
    test_send_msg()
    test_recv_msg()
    test_recv_msg_with_latencies()
    test_add_cvalue()
    test_add_cvalue_with_latencies()
    test_multiply_const()
    test_bit_cond_mult()
    test_run_subroutine()
    test_run_subroutine_2()
    test_run_request()
    test_return_result()



from __future__ import annotations

import math
import os
from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional, Type

import netsquid as ns
from netsquid.components import QuantumProcessor
from netsquid.qubits import ketstates, qubitapi

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo, UnitModule
from qoala.lang.hostlang import RunSubroutineOp
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.lhi import LhiLatencies, LhiLinkInfo, LhiTopology, LhiTopologyBuilder
from qoala.runtime.lhi_to_ehi import (
    GenericToVanillaInterface,
    LhiConverter,
    NativeToFlavourInterface,
)
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.entdist.entdist import EntDist
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.hostinterface import HostInterface
from qoala.sim.process import QoalaProcess
from qoala.sim.procnode import ProcNode
from qoala.util.math import has_state


def create_process(
    pid: int,
    program: QoalaProgram,
    unit_module: UnitModule,
    host_interface: HostInterface,
    inputs: Optional[Dict[str, Any]] = None,
) -> QoalaProcess:
    if inputs is None:
        inputs = {}
    prog_input = ProgramInput(values=inputs)
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=prog_input,
        unit_module=unit_module,
        block_tasks=[],
    )
    mem = ProgramMemory(pid=0)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={
            id: ClassicalSocket(host_interface, name)
            for (id, name) in program.meta.csockets.items()
        },
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def create_network_info(names: List[str]) -> NetworkInfo:
    return NetworkInfo.with_nodes({i: name for i, name in enumerate(names)})


def create_procnode(
    part: str,
    name: str,
    env: NetworkInfo,
    num_qubits: int,
    network_ehi: EhiNetworkInfo,
    procnode_cls: Type[ProcNode] = ProcNode,
    asynchronous: bool = False,
    pid: int = 0,
) -> ProcNode:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits)
    qprocessor = build_qprocessor_from_topology(f"{name}_processor", topology)

    node_id = env.get_node_id(name)
    procnode = procnode_cls(
        part=part,
        name=name,
        network_info=env,
        qprocessor=qprocessor,
        qdevice_topology=topology,
        latencies=LhiLatencies(qnos_instr_time=1000),
        ntf_interface=GenericToVanillaInterface(),
        network_ehi=network_ehi,
        node_id=node_id,
        asynchronous=asynchronous,
        pid=pid,
    )

    return procnode


class BqcProcNode(ProcNode):
    def __init__(
        self,
        name: str,
        network_info: NetworkInfo,
        qprocessor: QuantumProcessor,
        qdevice_topology: LhiTopology,
        latencies: LhiLatencies,
        ntf_interface: NativeToFlavourInterface,
        network_ehi: EhiNetworkInfo,
        node_id: Optional[int] = None,
        asynchronous: bool = False,
        pid: int = 0,
    ) -> None:
        super().__init__(
            name=name,
            network_info=network_info,
            qprocessor=qprocessor,
            qdevice_topology=qdevice_topology,
            latencies=latencies,
            ntf_interface=ntf_interface,
            network_ehi=network_ehi,
            node_id=node_id,
            asynchronous=asynchronous,
        )
        self.pid = pid

    def run_subroutine(
        self, process: QoalaProcess, host_instr_index: int, subrt_name: str
    ) -> Generator[EventExpression, None, None]:
        host_instr = process.program.instructions[host_instr_index]
        assert isinstance(host_instr, RunSubroutineOp)
        lrcall = self.host.processor.prepare_lr_call(process, host_instr)
        self.scheduler.qpudriver.allocate_qubits_for_routine(
            process, lrcall.routine_name
        )
        yield from self.qnos.processor.assign_local_routine(
            process, lrcall.routine_name, lrcall.input_addr, lrcall.result_addr
        )
        self.scheduler.qpudriver.free_qubits_after_routine(process, lrcall.routine_name)
        self.host.processor.post_lr_call(process, host_instr, lrcall)

    def run_request(
        self, process: QoalaProcess, host_instr_index: int, req_name: str
    ) -> Generator[EventExpression, None, None]:
        host_instr = process.program.instructions[host_instr_index]
        rrcall = self.host.processor.prepare_rr_call(process, host_instr)
        assert rrcall.routine_name == req_name
        yield from self.netstack.processor.assign_request_routine(process, rrcall)


class ServerProcNode(BqcProcNode):
    def __init__(self, part: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self._part = part

    def run(self) -> Generator[EventExpression, None, None]:
        self.finished = False

        process = self.memmgr.get_process(self.pid)
        self.scheduler.initialize_process(process)

        # csocket = assign_cval() : 0
        yield from self.host.processor.assign_instr_index(process, 0)
        # run_request() : req0
        yield from self.run_request(process, 1, "req0")
        # run_request() : req1
        yield from self.run_request(process, 2, "req1")

        if self._part == "only_rsp":
            self.finished = True
            return

        # run_subroutine(vec<client_id>) : local_cphase
        yield from self.run_subroutine(process, 3, "local_cphase")
        # delta1 = recv_cmsg(client_id)
        yield from self.host.processor.assign_instr_index(process, 4)
        # vec<m1> = run_subroutine(vec<delta1>) : meas_qubit_1
        yield from self.run_subroutine(process, 5, "meas_qubit_1")
        # send_cmsg(csocket, m1)
        yield from self.host.processor.assign_instr_index(process, 6)
        # delta2 = recv_cmsg(csocket)
        yield from self.host.processor.assign_instr_index(process, 7)

        if self._part == "meas_first_epr":
            self.finished = True
            return

        # vec<m2> = run_subroutine(vec<delta2>) : meas_qubit_0
        yield from self.run_subroutine(process, 8, "meas_qubit_0")
        # return_result(m1)
        yield from self.host.processor.assign_instr_index(process, 9)
        # return_result(m2)
        yield from self.host.processor.assign_instr_index(process, 10)

        assert self._part == "full"
        self.finished = True


class ClientProcNode(BqcProcNode):
    def __init__(self, part: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self._part = part

    def run(self) -> Generator[EventExpression, None, None]:
        self.finished = False

        process = self.memmgr.get_process(self.pid)
        self.scheduler.initialize_process(process)

        # csocket = assign_cval() : 0
        yield from self.host.processor.assign_instr_index(process, 0)
        # run_request() : req0
        yield from self.run_request(process, 1, "req0")
        # run_subroutine(vec<theta2>) : post_epr_0
        yield from self.run_subroutine(process, 2, "post_epr_0")
        # run_request() : req1
        yield from self.run_request(process, 3, "req1")
        # run_subroutine(vec<theta1>) : post_epr_1
        yield from self.run_subroutine(process, 4, "post_epr_1")
        # x = mult_const(p1) : 16
        # minus_theta1 = mult_const(theta1) : -1
        # delta1 = add_cval_c(minus_theta1, x)
        # delta1 = add_cval_c(delta1, alpha)
        # send_cmsg(server_id, delta1)
        for i in range(5, 10):
            yield from self.host.processor.assign_instr_index(process, i)

        if self._part == "only_rsp":
            self.finished = True
            return

        # m1 = recv_cmsg(csocket)
        yield from self.host.processor.assign_instr_index(process, 10)
        # y = mult_const(p2) : 16
        # minus_theta2 = mult_const(theta2) : -1
        # beta = bcond_mult_const(beta, m1) : -1
        # delta2 = add_cval_c(beta, minus_theta2)
        # delta2 = add_cval_c(delta2, y)
        # send_cmsg(csocket, delta2)
        for i in range(11, 17):
            yield from self.host.processor.assign_instr_index(process, i)

        if self._part == "meas_first_epr":
            self.finished = True
            return

        # return_result(p1)
        yield from self.host.processor.assign_instr_index(process, 17)
        # return_result(p2)
        yield from self.host.processor.assign_instr_index(process, 18)

        assert self._part == "full"
        self.finished = True


@dataclass
class BqcResult:
    client_process: QoalaProcess
    server_process: QoalaProcess
    client_procnode: BqcProcNode
    server_procnode: BqcProcNode


def run_bqc(
    part: str,
    alpha,
    beta,
    theta1,
    theta2,
    client_pid: int = 0,
    server_pid: int = 0,
):
    ns.sim_reset()

    num_qubits = 3
    network_info = create_network_info(names=["client", "server"])
    server_id = network_info.get_node_id("server")
    client_id = network_info.get_node_id("client")

    link_info = LhiLinkInfo.perfect(1000)
    ehi_link = LhiConverter.link_info_to_ehi(link_info)
    network_ehi = EhiNetworkInfo.fully_connected([server_id, client_id], ehi_link)

    path = os.path.join(os.path.dirname(__file__), "bqc_server.iqoala")
    with open(path) as file:
        server_text = file.read()
    server_program = QoalaParser(server_text).parse()

    server_procnode = create_procnode(
        part,
        "server",
        network_info,
        num_qubits,
        network_ehi,
        ServerProcNode,
        pid=server_pid,
    )
    server_ehi = server_procnode.memmgr.get_ehi()
    server_process = create_process(
        pid=server_pid,
        program=server_program,
        unit_module=UnitModule.from_full_ehi(server_ehi),
        host_interface=server_procnode.host._interface,
        inputs={"client_id": client_id},
    )
    server_procnode.add_process(server_process)

    path = os.path.join(os.path.dirname(__file__), "bqc_client.iqoala")
    with open(path) as file:
        client_text = file.read()
    client_program = QoalaParser(client_text).parse()

    client_procnode = create_procnode(
        part,
        "client",
        network_info,
        num_qubits,
        network_ehi,
        ClientProcNode,
        pid=client_pid,
    )
    client_ehi = client_procnode.memmgr.get_ehi()
    client_process = create_process(
        pid=client_pid,
        program=client_program,
        unit_module=UnitModule.from_full_ehi(client_ehi),
        host_interface=client_procnode.host._interface,
        inputs={
            "server_id": server_id,
            "alpha": alpha,
            "beta": beta,
            "theta1": theta1,
            "theta2": theta2,
        },
    )
    client_procnode.add_process(client_process)

    client_procnode.connect_to(server_procnode)

    nodes = [client_procnode.node, server_procnode.node]
    entdistcomp = EntDistComponent(network_info)
    client_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("client"))
    client_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("client"))
    server_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("server"))
    server_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("server"))
    entdist = EntDist(nodes=nodes, network_info=network_info, comp=entdistcomp)
    entdist.add_sampler(client_procnode.node.ID, server_procnode.node.ID, link_info)

    server_procnode.start()
    client_procnode.start()
    entdist.start()
    ns.sim_run()

    assert client_procnode.finished
    assert server_procnode.finished

    return BqcResult(
        client_process=client_process,
        server_process=server_process,
        client_procnode=client_procnode,
        server_procnode=server_procnode,
    )


def expected_rsp_state(theta: int, p: int, dummy: bool):
    expected = qubitapi.create_qubits(1)[0]

    if dummy:
        if p == 0:
            return ketstates.s0
        elif p == 1:
            return ketstates.s1
    else:
        if (theta, p) == (0, 0):
            return ketstates.h0
        elif (theta, p) == (0, 1):
            return ketstates.h1
        if (theta, p) == (8, 0):
            return ketstates.y0
        elif (theta, p) == (8, 1):
            return ketstates.y1
        if (theta, p) == (16, 0):
            return ketstates.h1
        elif (theta, p) == (16, 1):
            return ketstates.h0
        if (theta, p) == (-8, 0):
            return ketstates.y1
        elif (theta, p) == (-8, 1):
            return ketstates.y0

    return expected.qstate


def test_bqc_only_rsp():
    ns.sim_reset()

    alpha_beta_theta1_theta2 = [
        (0, 0, 0, 0),
        (0, 8, 0, 0),
        (8, 0, 0, 0),
        (8, 8, 0, 0),
    ]

    for alpha, beta, theta1, theta2 in alpha_beta_theta1_theta2:
        for _ in range(10):
            result = run_bqc("only_rsp", alpha, beta, theta1, theta2)

            p1 = result.client_process.host_mem.read("p1")
            p2 = result.client_process.host_mem.read("p2")
            delta1 = result.client_process.host_mem.read("delta1")
            q0 = result.server_procnode.qdevice.get_local_qubit(0)
            q1 = result.server_procnode.qdevice.get_local_qubit(1)

            assert delta1 == alpha - theta1 + p1 * 16

            # p2 and theta2 control state of q0
            expected_q0 = expected_rsp_state(theta2, p2, dummy=False)
            assert has_state(q0, expected_q0)

            # p1 and theta1 control state of q1
            expected_q1 = expected_rsp_state(theta1, p1, dummy=False)
            assert has_state(q1, expected_q1)


def test_bqc_only_meas_first_EPR():
    ns.sim_reset()

    alpha_beta_theta1_theta2 = [
        (0, 0, 0, 0),
        (0, 8, 0, 0),
        (8, 0, 0, 0),
        (8, 8, 0, 0),
    ]

    for alpha, beta, theta1, theta2 in alpha_beta_theta1_theta2:
        for _ in range(10):
            result = run_bqc("meas_first_epr", alpha, beta, theta1, theta2)

            p1 = result.client_process.host_mem.read("p1")
            p2 = result.client_process.host_mem.read("p2")
            delta1 = result.client_process.host_mem.read("delta1")
            m1 = result.client_process.host_mem.read("m1")
            delta2 = result.client_process.host_mem.read("delta2")

            q0 = result.server_procnode.qdevice.get_local_qubit(0)
            print(q0.qstate)

            assert delta1 == alpha - theta1 + p1 * 16
            assert delta2 == math.pow(-1, m1) * beta - theta2 + p2 * 16


def test_bqc_full():
    # Effective computation: measure in Z the following state:
    # H Rz(beta) H Rz(alpha) |+>
    # m2 should be this outcome

    # angles are in multiples of pi/16

    def check(alpha, beta, theta1, theta2, expected):
        ns.sim_reset()

        results = [
            run_bqc("full", alpha, beta, theta1, theta2, client_pid=i, server_pid=i)
            for i in range(2)
        ]
        m2s = [result.server_process.result.values["m2"] for result in results]
        assert all(m2 == expected for m2 in m2s)

    check(alpha=8, beta=8, theta1=0, theta2=0, expected=0)
    check(alpha=8, beta=24, theta1=0, theta2=0, expected=1)
    check(alpha=8, beta=8, theta1=13, theta2=27, expected=0)
    check(alpha=8, beta=24, theta1=2, theta2=22, expected=1)


if __name__ == "__main__":
    test_bqc_only_rsp()
    test_bqc_only_meas_first_EPR()
    test_bqc_full()



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional

from netqasm.lang.subroutine import Subroutine


@dataclass
class RoutineMetadata:
    # IDs in unit module of virtual qubits that are
    # used in this routine
    qubit_use: List[int]

    # IDs in unit module of virtual qubits that still have a state
    # that should be kept after finishing this routine
    qubit_keep: List[int]

    @classmethod
    def use_none(cls) -> RoutineMetadata:
        return RoutineMetadata([], [])

    @classmethod
    def free_all(cls, ids: List[int]) -> RoutineMetadata:
        return RoutineMetadata(ids, [])


class LocalRoutine:
    def __init__(
        self,
        name: str,
        subrt: Subroutine,
        return_vars: List[str],
        metadata: RoutineMetadata,
        request_name: Optional[str] = None,
    ) -> None:
        self._name = name
        self._subrt = subrt
        self._return_vars = return_vars
        self._metadata = metadata
        self._request_name = request_name

    @property
    def name(self) -> str:
        return self._name

    @property
    def subroutine(self) -> Subroutine:
        return self._subrt

    @property
    def return_vars(self) -> List[str]:
        return self._return_vars

    @property
    def metadata(self) -> RoutineMetadata:
        return self._metadata

    @property
    def request_name(self) -> Optional[str]:
        return self._request_name

    def serialize(self) -> str:
        s = f"SUBROUTINE {self.name}"
        s += f"\nparams: {', '.join(self.subroutine.arguments)}"
        s += f"\nreturns: {', '.join(str(v) for v in self.return_vars)}"
        s += f"\nuses: {', '.join(str(q) for q in self.metadata.qubit_use)}"
        s += f"\nkeeps: {', '.join(str(q) for q in self.metadata.qubit_keep)}"
        s += "\nNETQASM_START\n"
        s += self.subroutine.print_instructions()
        s += "\nNETQASM_END"
        return s

    def __str__(self) -> str:
        s = "\n"
        for value in self.return_vars:
            s += f"return {str(value)}\n"
        s += "NETQASM_START\n"
        s += self.subroutine.print_instructions()
        s += "\nNETQASM_END"
        return s

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LocalRoutine):
            return NotImplemented
        return (
            self.name == other.name
            and self.subroutine == other.subroutine
            and self.metadata == other.metadata
            and self.return_vars == other.return_vars
        )



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Dict, List

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.program import BatchInfo, BatchResult, ProgramInput
from qoala.runtime.schedule import TaskSchedule
from qoala.sim.build import build_network


def create_network_info(names: List[str]) -> NetworkInfo:
    env = NetworkInfo.with_nodes({i: name for i, name in enumerate(names)})
    env.set_global_schedule([0, 1, 2])
    env.set_timeslot_len(1e6)
    return env


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def create_batch(
    program: QoalaProgram,
    unit_module: UnitModule,
    inputs: List[ProgramInput],
    num_iterations: int,
) -> BatchInfo:
    return BatchInfo(
        program=program,
        unit_module=unit_module,
        inputs=inputs,
        num_iterations=num_iterations,
        deadline=0,
    )


@dataclass
class BqcResult:
    client_results: Dict[int, BatchResult]
    server_results: Dict[int, BatchResult]


def run_bqc(alpha, beta, theta1, theta2, num_iterations: int) -> BqcResult:
    ns.sim_reset()

    num_qubits = 3
    network_info = create_network_info(names=["client", "server"])
    server_id = network_info.get_node_id("server")
    client_id = network_info.get_node_id("client")

    server_node_cfg = create_procnode_cfg("server", server_id, num_qubits)
    client_node_cfg = create_procnode_cfg("client", client_id, num_qubits)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[server_node_cfg, client_node_cfg], link_duration=1000
    )
    network = build_network(network_cfg, network_info)
    server_procnode = network.nodes["server"]
    client_procnode = network.nodes["client"]

    server_program = load_program("bqc_server.iqoala")
    server_inputs = [
        ProgramInput({"client_id": client_id}) for _ in range(num_iterations)
    ]

    server_unit_module = UnitModule.from_full_ehi(server_procnode.memmgr.get_ehi())
    server_batch = create_batch(
        server_program, server_unit_module, server_inputs, num_iterations
    )
    server_procnode.submit_batch(server_batch)
    server_procnode.initialize_processes()
    server_tasks = server_procnode.scheduler.get_tasks_to_schedule()
    server_schedule = TaskSchedule.consecutive(server_tasks)
    server_procnode.scheduler.upload_schedule(server_schedule)

    client_program = load_program("bqc_client.iqoala")
    client_inputs = [
        ProgramInput(
            {
                "server_id": server_id,
                "alpha": alpha,
                "beta": beta,
                "theta1": theta1,
                "theta2": theta2,
            }
        )
        for _ in range(num_iterations)
    ]

    client_unit_module = UnitModule.from_full_ehi(client_procnode.memmgr.get_ehi())
    client_batch = create_batch(
        client_program, client_unit_module, client_inputs, num_iterations
    )
    client_procnode.submit_batch(client_batch)
    client_procnode.initialize_processes()
    client_tasks = client_procnode.scheduler.get_tasks_to_schedule()
    client_schedule = TaskSchedule.consecutive(client_tasks)
    client_procnode.scheduler.upload_schedule(client_schedule)

    network.start()
    ns.sim_run()

    client_results = client_procnode.scheduler.get_batch_results()
    server_results = server_procnode.scheduler.get_batch_results()

    return BqcResult(client_results, server_results)


def test_bqc():
    # Effective computation: measure in Z the following state:
    # H Rz(beta) H Rz(alpha) |+>
    # m2 should be this outcome

    # angles are in multiples of pi/16

    # LogManager.set_log_level("DEBUG")
    # LogManager.log_to_file("test_run.log")

    def check(alpha, beta, theta1, theta2, expected, num_iterations):
        ns.sim_reset()
        bqc_result = run_bqc(
            alpha=alpha,
            beta=beta,
            theta1=theta1,
            theta2=theta2,
            num_iterations=num_iterations,
        )
        assert len(bqc_result.client_results) > 0
        assert len(bqc_result.server_results) > 0

        server_batch_results = bqc_result.server_results
        for _, batch_results in server_batch_results.items():
            program_results = batch_results.results
            m2s = [result.values["m2"] for result in program_results]
            assert all(m2 == expected for m2 in m2s)

    check(alpha=8, beta=8, theta1=0, theta2=0, expected=0, num_iterations=10)
    check(alpha=8, beta=24, theta1=0, theta2=0, expected=1, num_iterations=10)
    check(alpha=8, beta=8, theta1=13, theta2=27, expected=0, num_iterations=10)
    check(alpha=8, beta=24, theta1=2, theta2=22, expected=1, num_iterations=10)


if __name__ == "__main__":
    test_bqc()



from __future__ import annotations

from qoala.runtime.environment import NetworkInfo
from qoala.sim.entdist.entdistcomp import EntDistComponent


def create_entdistcomp(num_nodes: int) -> EntDistComponent:
    nodes = {id: f"node_{id}" for id in range(num_nodes)}
    env = NetworkInfo.with_nodes(nodes)

    return EntDistComponent(env)


def test_one_node():
    comp = create_entdistcomp(num_nodes=1)

    # should have 2 node ports
    assert len(comp.ports) == 2
    assert "node_node_0_in" in comp.ports
    assert "node_node_0_out" in comp.ports

    # Test properties
    assert comp.node_in_port("node_0") == comp.ports["node_node_0_in"]
    assert comp.node_out_port("node_0") == comp.ports["node_node_0_out"]


def test_many_nodes():
    comp = create_entdistcomp(num_nodes=5)

    # should have 5 * 2 node ports
    assert len(comp.ports) == 10

    for i in range(5):
        assert f"node_node_{i}_in" in comp.ports
        assert f"node_node_{i}_out" in comp.ports
        # Test properties
        assert comp.node_in_port(f"node_{i}") == comp.ports[f"node_node_{i}_in"]
        assert comp.node_out_port(f"node_{i}") == comp.ports[f"node_node_{i}_out"]


if __name__ == "__main__":
    test_one_node()
    test_many_nodes()



import logging
from copy import deepcopy
from typing import Any, Dict, Generator, List, Optional

from pydynaa import EventExpression
from qoala.lang.request import CallbackType, EprType, QoalaRequest
from qoala.runtime.lhi import INSTR_MEASURE_INSTANT
from qoala.runtime.memory import ProgramMemory, RunningRequestRoutine
from qoala.runtime.message import Message, RrCallTuple
from qoala.sim.entdist.entdist import EntDistRequest
from qoala.sim.events import MSG_REQUEST_DELIVERED
from qoala.sim.netstack.netstackinterface import NetstackInterface, NetstackLatencies
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice, QDeviceCommand
from qoala.sim.qnos.qnosprocessor import QnosProcessor
from qoala.util.logging import LogManager


class NetstackProcessor:
    def __init__(
        self, interface: NetstackInterface, latencies: NetstackLatencies
    ) -> None:
        self._interface = interface
        self._latencies = latencies

        self._name = f"{interface.name}_NetstackProcessor"

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._name})"
        )

        # memory of current program, only not-None when processor is active
        self._current_prog_mem: Optional[ProgramMemory] = None

        self._current_routine: Optional[RunningRequestRoutine] = None

    def _prog_mem(self) -> ProgramMemory:
        # May only be called when processor is active
        assert self._current_prog_mem is not None
        return self._current_prog_mem

    def _routine(self) -> RunningRequestRoutine:
        # May only be called when processor is active
        assert self._current_routine is not None
        return self._current_routine

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    def execute_entdist_request(
        self, request: EntDistRequest
    ) -> Generator[EventExpression, None, None]:
        self._interface.send_entdist_msg(Message(request))
        result = yield from self._interface.receive_entdist_msg()
        if result.content != MSG_REQUEST_DELIVERED:
            raise RuntimeError("Request was not served")

    def allocate_for_pair(
        self, process: QoalaProcess, request: QoalaRequest, index: int
    ) -> int:
        memmgr = self._interface.memmgr
        pid = process.pid

        virt_id = request.virt_ids.get_id(index)
        memmgr.allocate(pid, virt_id)

        return virt_id

    def create_entdist_request(
        self, process: QoalaProcess, request: QoalaRequest, virt_id: int
    ) -> EntDistRequest:
        memmgr = self._interface.memmgr
        pid = process.pid
        phys_id = memmgr.phys_id_for(pid, virt_id)

        return EntDistRequest(
            local_node_id=self._interface.node_id,
            remote_node_id=request.remote_id,
            local_qubit_id=phys_id,
        )

    def measure_epr_qubit(
        self, process: QoalaProcess, virt_id: int
    ) -> Generator[EventExpression, None, int]:
        phys_id = self._interface.memmgr.phys_id_for(process.pid, virt_id)
        # Should have been allocated by `handle_req_routine_md`
        assert phys_id is not None
        # Use the special INSTR_MEASURE_INSTANT instruction so that measuring
        # doesn't take time.
        commands = [QDeviceCommand(INSTR_MEASURE_INSTANT, [phys_id])]
        m = yield from self.qdevice.execute_commands(commands=commands)
        assert m is not None
        return m

    def handle_req_routine_md(
        self, process: QoalaProcess, routine_name: str
    ) -> Generator[EventExpression, None, None]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine
        request = routine.request
        assert request.typ == EprType.MEASURE_DIRECTLY
        num_pairs = request.num_pairs

        outcomes: List[int] = []

        if routine.callback_type == CallbackType.SEQUENTIAL:
            raise NotImplementedError
        else:
            for i in range(num_pairs):
                virt_id = self.allocate_for_pair(process, request, i)
                entdist_req = self.create_entdist_request(process, request, virt_id)
                # Create EPR pair
                yield from self.execute_entdist_request(entdist_req)
                # Measure local qubit
                m = yield from self.measure_epr_qubit(process, virt_id)
                # Free virt qubit
                self._interface.memmgr.free(process.pid, virt_id)
                outcomes.append(m)

        shared_mem = process.prog_memory.shared_mem
        results_addr = running_routine.result_addr
        shared_mem.write_rr_out(results_addr, outcomes)

    def handle_req_routine_ck(
        self,
        process: QoalaProcess,
        routine_name: str,
        qnosprocessor: Optional[QnosProcessor] = None,
    ) -> Generator[EventExpression, None, None]:
        running_routine = process.qnos_mem.get_running_request_routine(routine_name)
        routine = running_routine.routine
        request = routine.request
        num_pairs = request.num_pairs

        if routine.callback_type == CallbackType.SEQUENTIAL:
            for i in range(num_pairs):
                virt_id = self.allocate_for_pair(process, request, i)
                entdist_req = self.create_entdist_request(process, request, virt_id)
                yield from self.execute_entdist_request(entdist_req)

                if routine.callback is not None:
                    # TODO: write CR inputs to shared memory

                    # Allocate qubits for CR
                    cb_routine = process.get_local_routine(routine.callback)
                    for virt_id in cb_routine.metadata.qubit_use:
                        if (
                            self._interface.memmgr.phys_id_for(process.pid, virt_id)
                            is None
                        ):
                            self._interface.memmgr.allocate(process.pid, virt_id)

                    assert qnosprocessor is not None
                    yield from qnosprocessor.assign_local_routine(
                        process=process,
                        routine_name=routine.callback,
                        input_addr=running_routine.cb_input_addrs[i],
                        result_addr=running_routine.cb_output_addrs[i],
                    )

                    # Free CR qubits
                    for virt_id in cb_routine.metadata.qubit_use:
                        if virt_id not in cb_routine.metadata.qubit_keep:
                            self._interface.memmgr.free(process.pid, virt_id)
        else:
            for i in range(num_pairs):
                virt_id = self.allocate_for_pair(process, request, i)
                entdist_req = self.create_entdist_request(process, request, virt_id)
                yield from self.execute_entdist_request(entdist_req)

            if routine.callback is not None:
                # TODO: write CR inputs to shared memory

                # Allocate qubits for CR
                cb_routine = process.get_local_routine(routine.callback)
                for virt_id in cb_routine.metadata.qubit_use:
                    if self._interface.memmgr.phys_id_for(process.pid, virt_id) is None:
                        self._interface.memmgr.allocate(process.pid, virt_id)

                # for WAIT_ALL, there should be at most 1 callback.
                # So we can access index 0 of the cb input/output addresses.

                assert qnosprocessor is not None
                yield from qnosprocessor.assign_local_routine(
                    process=process,
                    routine_name=routine.callback,
                    input_addr=running_routine.cb_input_addrs[0],
                    result_addr=running_routine.cb_output_addrs[0],
                )

                # Free CR qubits
                for virt_id in cb_routine.metadata.qubit_use:
                    if virt_id not in cb_routine.metadata.qubit_keep:
                        self._interface.memmgr.free(process.pid, virt_id)

    def instantiate_routine(
        self,
        process: QoalaProcess,
        rrcall: RrCallTuple,
        args: Dict[str, Any],
    ) -> None:
        """Instantiates and activates routine."""
        routine = process.get_request_routine(rrcall.routine_name)
        instance = deepcopy(routine)
        instance.request.instantiate(args)

        running_routine = RunningRequestRoutine(
            instance,
            rrcall.input_addr,
            rrcall.result_addr,
            rrcall.cb_input_addrs,
            rrcall.cb_output_addrs,
        )
        process.qnos_mem.add_running_request_routine(running_routine)

    def assign_request_routine(
        self,
        process: QoalaProcess,
        rrcall: RrCallTuple,
        qnosprocessor: Optional[QnosProcessor] = None,
    ) -> Generator[EventExpression, None, None]:
        routine = process.get_request_routine(rrcall.routine_name)
        global_args = process.prog_instance.inputs.values
        self.instantiate_routine(process, rrcall, global_args)

        if routine.request.typ == EprType.CREATE_KEEP:
            yield from self.handle_req_routine_ck(
                process, rrcall.routine_name, qnosprocessor
            )
        elif routine.request.typ == EprType.MEASURE_DIRECTLY:
            yield from self.handle_req_routine_md(process, rrcall.routine_name)
        else:
            raise NotImplementedError



from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

from netqasm.lang.instr.flavour import Flavour, VanillaFlavour
from netqasm.lang.operand import Template
from netqasm.lang.parsing.text import parse_text_subroutine

from qoala.lang import hostlang as hl
from qoala.lang.program import LocalRoutine, ProgramMeta, QoalaProgram
from qoala.lang.request import (
    CallbackType,
    EprRole,
    EprType,
    QoalaRequest,
    RequestRoutine,
    RequestVirtIdMapping,
)
from qoala.lang.routine import RoutineMetadata

LHR_OP_NAMES: Dict[str, hl.ClassicalIqoalaOp] = {
    cls.OP_NAME: cls  # type: ignore
    for cls in [
        hl.SendCMsgOp,
        hl.ReceiveCMsgOp,
        hl.AddCValueOp,
        hl.MultiplyConstantCValueOp,
        hl.BitConditionalMultiplyConstantCValueOp,
        hl.AssignCValueOp,
        hl.RunSubroutineOp,
        hl.RunRequestOp,
        hl.ReturnResultOp,
    ]
}


class EndOfTextException(Exception):
    pass


class QoalaParseError(Exception):
    pass


class IqoalaMetaParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_meta_line(self, key: str, line: str) -> List[str]:
        split = line.split(":")
        assert len(split) >= 1
        assert split[0] == key
        if len(split) == 1:
            return []
        assert len(split) == 2
        if len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_meta_mapping(self, values: List[str]) -> Dict[int, str]:
        result_dict = {}
        for v in values:
            key_value = [x.strip() for x in v.split("->")]
            assert len(key_value) == 2
            result_dict[int(key_value[0].strip())] = key_value[1].strip()
        return result_dict

    def parse(self) -> ProgramMeta:
        try:
            start_line = self._read_line()
            assert start_line == "META_START"

            name_values = self._parse_meta_line("name", self._read_line())
            assert len(name_values) == 1
            name = name_values[0]

            parameters = self._parse_meta_line("parameters", self._read_line())

            csockets_map = self._parse_meta_line("csockets", self._read_line())
            csockets = self._parse_meta_mapping(csockets_map)
            epr_sockets_map = self._parse_meta_line("epr_sockets", self._read_line())
            epr_sockets = self._parse_meta_mapping(epr_sockets_map)

            end_line = self._read_line()
            if end_line != "META_END":
                raise QoalaParseError("Could not parse meta.")
        except AssertionError:
            raise QoalaParseError
        except EndOfTextException:
            raise QoalaParseError

        return ProgramMeta(name, parameters, csockets, epr_sockets)


class IqoalaInstrParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_var(self, var_str: str) -> Union[str, hl.IqoalaVector]:
        if var_str.startswith("vec<"):
            vec_values_str = var_str[4:-1]
            if len(vec_values_str) == 0:
                vec_values = []
            else:
                vec_values = [x.strip() for x in vec_values_str.split(";")]
            return hl.IqoalaVector(vec_values)
        else:
            return var_str

    def _parse_lhr(self) -> hl.ClassicalIqoalaOp:
        line = self._read_line()

        attr: Optional[hl.IqoalaValue]

        assign_parts = [x.strip() for x in line.split("=")]
        assert len(assign_parts) <= 2
        if len(assign_parts) == 1:
            value = assign_parts[0]
            result = None
        elif len(assign_parts) == 2:
            value = assign_parts[1]
            result = self._parse_var(assign_parts[0])
        value_parts = [x.strip() for x in value.split(":")]
        assert len(value_parts) <= 2
        if len(value_parts) == 2:
            value = value_parts[0]
            attr_str = value_parts[1]
            try:
                attr = int(attr_str)
            except ValueError:
                attr = attr_str
        else:
            value = value_parts[0]
            attr = None

        op_parts = [x.strip() for x in value.split("(")]
        assert len(op_parts) == 2
        op = op_parts[0]
        arguments = op_parts[1].rstrip(")")
        if len(arguments) == 0:
            raw_args = []
        else:
            raw_args = [x.strip() for x in arguments.split(",")]

        args = [self._parse_var(arg) for arg in raw_args]

        # print(f"result = {result}, op = {op}, args = {args}, attr = {attr}")

        lhr_op = LHR_OP_NAMES[op].from_generic_args(result, args, attr)  # type: ignore
        return lhr_op

    def parse(self) -> List[hl.ClassicalIqoalaOp]:
        instructions: List[hl.ClassicalIqoalaOp] = []

        try:
            while True:
                instr = self._parse_lhr()
                instructions.append(instr)
        except AssertionError:
            raise QoalaParseError
        except EndOfTextException:
            pass

        return instructions


class HostCodeParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def get_block_texts(self) -> List[str]:
        block_start_lines: List[int] = []

        for i, line in enumerate(self._lines):
            if line.startswith("^"):
                block_start_lines.append(i)

        assert len(block_start_lines) > 0

        block_texts: List[str] = []
        for i in range(len(block_start_lines) - 1):
            start = block_start_lines[i]
            end = block_start_lines[i + 1]
            text = self._lines[start:end]
            block_texts.append("\n".join([line for line in text]))

        last = block_start_lines[-1]
        last_text = self._lines[last:]
        block_texts.append("\n".join([line for line in last_text]))

        return block_texts

    def _parse_block_header(self, line: str) -> Tuple[str, hl.BasicBlockType]:
        # return (block name, block type)
        assert line.startswith("^")
        split_space = line.split(" ")
        assert len(split_space) > 0
        name = split_space[0][1:]  # trim '^' at start
        typ_index = line.find("type =")
        assert typ_index >= 0
        typ_end_index = line.find("}")
        assert typ_end_index >= 0
        raw_typ = line[typ_index + 7 : typ_end_index]
        typ = hl.BasicBlockType[raw_typ.upper()]
        return name, typ

    def parse_block(self, text: str) -> hl.BasicBlock:
        lines = [line.strip() for line in text.split("\n")]
        lines = [line for line in lines if len(line) > 0]
        name, typ = self._parse_block_header(lines[0])
        instr_lines = lines[1:]
        instrs = IqoalaInstrParser("\n".join(instr_lines)).parse()

        return hl.BasicBlock(name, typ, instrs)

    def parse(self) -> List[hl.BasicBlock]:
        block_texts = self.get_block_texts()
        return [self.parse_block(text) for text in block_texts]


class LocalRoutineParser:
    def __init__(self, text: str, flavour: Optional[Flavour] = None) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0
        if flavour is None:
            flavour = VanillaFlavour()
        self._flavour = flavour

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_subrt_meta_line(self, key: str, line: str) -> List[str]:
        split = line.split(":")
        assert len(split) >= 1
        assert split[0] == key
        if len(split) == 1:
            return []
        assert len(split) == 2
        if len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_subroutine(self) -> LocalRoutine:
        name_line = self._read_line()
        assert name_line.startswith("SUBROUTINE ")
        name = name_line[len("SUBROUTINE") + 1 :]
        params_line = self._parse_subrt_meta_line("params", self._read_line())
        # TODO: use params line?

        return_vars = self._parse_subrt_meta_line("returns", self._read_line())
        assert all(" " not in v for v in return_vars)

        uses_line = self._parse_subrt_meta_line("uses", self._read_line())
        uses = [int(u) for u in uses_line]
        keeps_line = self._parse_subrt_meta_line("keeps", self._read_line())
        keeps = [int(k) for k in keeps_line]
        metadata = RoutineMetadata(qubit_use=uses, qubit_keep=keeps)

        request_line = self._parse_subrt_meta_line("request", self._read_line())
        assert len(request_line) in [0, 1]
        request_name = None if len(request_line) == 0 else request_line[0]

        start_line = self._read_line()
        assert start_line == "NETQASM_START"
        subrt_lines = []
        while True:
            line = self._read_line()
            if line == "NETQASM_END":
                break
            subrt_lines.append(line)
        subrt_text = "\n".join(subrt_lines)

        subrt = parse_text_subroutine(subrt_text, flavour=self._flavour)

        # Check that all templates are declared as params to the subroutine
        if any(arg not in params_line for arg in subrt.arguments):
            raise QoalaParseError
        return LocalRoutine(name, subrt, return_vars, metadata, request_name)

    def parse(self) -> Dict[str, LocalRoutine]:
        subroutines: Dict[str, LocalRoutine] = {}
        try:
            while True:
                try:
                    subrt = self._parse_subroutine()
                    subroutines[subrt.name] = subrt
                except AssertionError:
                    raise QoalaParseError
        except EndOfTextException:
            return subroutines


class RequestRoutineParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_request_line(self, key: str, line: str) -> List[str]:
        split = line.split(":")
        assert len(split) >= 1
        assert split[0] == key
        if len(split) == 1:
            return []
        assert len(split) == 2
        if len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_single_int_value(
        self, key: str, line: str, allow_template: bool = False
    ) -> Union[int, Template]:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise QoalaParseError
        value = strings[0]
        if allow_template:
            if value.startswith("{") and value.endswith("}"):
                value = value.strip("{}").strip()
                return Template(value)
        return int(value)

    def _parse_optional_str_value(self, key: str, line: str) -> Optional[str]:
        strings = self._parse_request_line(key, line)
        if len(strings) == 0:
            return None
        assert len(strings) == 1
        return strings[0]

    def _parse_int_list_value(self, key: str, line: str) -> List[int]:
        strings = self._parse_request_line(key, line)
        return [int(s) for s in strings]

    def _parse_single_float_value(
        self, key: str, line: str, allow_template: bool = False
    ) -> Union[float, Template]:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise QoalaParseError
        value = strings[0]
        if allow_template:
            if value.startswith("{") and value.endswith("}"):
                value = value.strip("{}").strip()
                return Template(value)
        return float(value)

    def _parse_epr_create_role_value(self, key: str, line: str) -> EprRole:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise QoalaParseError
        try:
            return EprRole[strings[0].upper()]
        except KeyError:
            raise QoalaParseError

    def _parse_epr_create_type_value(self, key: str, line: str) -> EprType:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise QoalaParseError
        try:
            return EprType[strings[0].upper()]
        except KeyError:
            raise QoalaParseError

    def _parse_virt_ids(self, key: str, line: str) -> RequestVirtIdMapping:
        split = line.split(":")
        assert len(split) >= 1
        assert split[0] == key
        value = split[1].strip()
        return RequestVirtIdMapping.from_str(value)

    def _parse_request(self) -> RequestRoutine:
        name_line = self._read_line()
        if not name_line.startswith("REQUEST "):
            raise QoalaParseError
        name = name_line[len("REQUEST") + 1 :]

        raw_callback_type = self._parse_request_line("callback_type", self._read_line())
        assert len(raw_callback_type) == 1
        callback_type = CallbackType[raw_callback_type[0].upper()]

        callback = self._parse_optional_str_value("callback", self._read_line())

        return_vars = self._parse_request_line("return_vars", self._read_line())

        remote_id = self._parse_single_int_value(
            "remote_id", self._read_line(), allow_template=True
        )
        epr_socket_id = self._parse_single_int_value(
            "epr_socket_id", self._read_line(), allow_template=True
        )
        num_pairs = self._parse_single_int_value(
            "num_pairs", self._read_line(), allow_template=True
        )
        # virt_ids = self._parse_int_list_value("virt_ids", self._read_line())
        virt_ids = self._parse_virt_ids("virt_ids", self._read_line())
        timeout = self._parse_single_int_value(
            "timeout", self._read_line(), allow_template=True
        )
        fidelity = self._parse_single_float_value(
            "fidelity", self._read_line(), allow_template=True
        )
        typ = self._parse_epr_create_type_value("typ", self._read_line())
        role = self._parse_epr_create_role_value("role", self._read_line())

        request = QoalaRequest(
            name=name,
            remote_id=remote_id,
            epr_socket_id=epr_socket_id,
            num_pairs=num_pairs,
            virt_ids=virt_ids,
            timeout=timeout,
            fidelity=fidelity,
            typ=typ,
            role=role,
        )
        return RequestRoutine(name, request, return_vars, callback_type, callback)

    def parse(self) -> Dict[str, RequestRoutine]:
        requests: Dict[str, RequestRoutine] = {}
        try:
            while True:
                request = self._parse_request()
                requests[request.name] = request
        except EndOfTextException:
            return requests


class QoalaParser:
    def __init__(
        self,
        text: Optional[str] = None,
        meta_text: Optional[str] = None,
        host_text: Optional[str] = None,
        subrt_text: Optional[str] = None,
        req_text: Optional[str] = None,
        flavour: Optional[Flavour] = None,
    ) -> None:
        if text is not None:
            meta_text, host_text, subrt_text, req_text = self._split_text(text)
        else:
            assert meta_text is not None
            assert host_text is not None
            assert subrt_text is not None
            assert req_text is not None
        self._meta_text = meta_text
        self._host_text = host_text
        self._subrt_text = subrt_text
        self._req_text = req_text
        self._meta_parser = IqoalaMetaParser(meta_text)
        self._host_parser = HostCodeParser(host_text)
        self._subrt_parser = LocalRoutineParser(subrt_text, flavour)
        self._req_parser = RequestRoutineParser(req_text)

    def _split_text(self, text: str) -> Tuple[str, str, str, str]:
        lines = [line.strip() for line in text.split("\n")]
        meta_end_line: int
        first_subrt_line: Optional[int] = None
        first_req_line: Optional[int] = None
        for i, line in enumerate(lines):
            if "META_END" in line:
                meta_end_line = i
                break
        for i, line in enumerate(lines):
            if "SUBROUTINE" in line:
                first_subrt_line = i
                break
        for i, line in enumerate(lines):
            if "REQUEST" in line:
                first_req_line = i
                break

        meta_text = "\n".join(lines[0 : meta_end_line + 1])
        host_end_line: Optional[int] = None
        if first_subrt_line is None and first_req_line is None:
            # no subroutines and no requests
            subrt_text = ""
            req_text = ""
        elif first_subrt_line is not None and first_req_line is None:
            # subroutines but no requests
            subrt_text = "\n".join(lines[first_subrt_line:])
            req_text = ""
            host_end_line = first_subrt_line
        elif first_subrt_line is None and first_req_line is not None:
            # no subroutines but only requests
            subrt_text = ""
            req_text = "\n".join(lines[first_req_line:])
            host_end_line = first_req_line
        else:
            # subroutines and requests
            subrt_text = "\n".join(lines[first_subrt_line:first_req_line])
            req_text = "\n".join(lines[first_req_line:])
            host_end_line = first_subrt_line
        if host_end_line is not None:
            host_text = "\n".join(lines[meta_end_line + 1 : host_end_line])
        else:
            host_text = "\n".join(lines[meta_end_line + 1 :])

        return meta_text, host_text, subrt_text, req_text

    def parse(self) -> QoalaProgram:
        blocks = self._host_parser.parse()
        subroutines = self._subrt_parser.parse()
        requests = self._req_parser.parse()
        meta = self._meta_parser.parse()

        # Check that all references to subroutines (in RunSubroutineOp instructions)
        # and to requests (in RunRequestOp instructions) are valid.
        for block in blocks:
            for instr in block.instructions:
                if isinstance(instr, hl.RunSubroutineOp):
                    subrt_name = instr.subroutine
                    if subrt_name not in subroutines:
                        raise QoalaParseError
                elif isinstance(instr, hl.RunRequestOp):
                    req_name = instr.req_routine
                    if req_name not in requests:
                        raise QoalaParseError

        return QoalaProgram(meta, blocks, subroutines, requests)



import itertools
from typing import List, Tuple

import netsquid as ns
from netsquid.nodes import Node

from qoala.runtime.environment import NetworkInfo
from qoala.runtime.lhi import LhiLinkInfo, LhiTopologyBuilder
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.entdist.entdist import EntDist, EntDistRequest
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.qdevice import QDevice
from qoala.util.math import B00_DENS, has_multi_state
from qoala.util.tests import netsquid_run


def create_n_qdevices(n: int, num_qubits: int = 1) -> List[QDevice]:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits)
    qdevices: List[QDevice] = []
    for i in range(n):
        qproc = build_qprocessor_from_topology(name=f"qproc_{i}", topology=topology)
        node = Node(name=f"node_{i}", qmemory=qproc)
        qdevices.append(QDevice(node=node, topology=topology))

    return qdevices


def create_entdist(qdevices: List[QDevice]) -> EntDist:
    env = NetworkInfo.with_nodes(
        {qdevice.node.ID: qdevice.node.name for qdevice in qdevices}
    )
    comp = EntDistComponent(env)
    entdist = EntDist(
        nodes=[qdevice.node for qdevice in qdevices], network_info=env, comp=comp
    )

    link_info = LhiLinkInfo.perfect(1000)
    for qd1, qd2 in itertools.combinations(qdevices, 2):
        entdist.add_sampler(qd1.node.ID, qd2.node.ID, link_info)

    return entdist


def create_request(
    node1_id: int, node2_id: int, local_qubit_id: int = 0
) -> EntDistRequest:
    return EntDistRequest(
        local_node_id=node1_id, remote_node_id=node2_id, local_qubit_id=local_qubit_id
    )


def create_request_pair(
    node1_id: int, node2_id: int, node1_qubit_id: int = 0, node2_qubit_id: int = 0
) -> Tuple[EntDistRequest]:
    req1 = EntDistRequest(
        local_node_id=node1_id, remote_node_id=node2_id, local_qubit_id=node1_qubit_id
    )
    req2 = EntDistRequest(
        local_node_id=node2_id, remote_node_id=node1_id, local_qubit_id=node2_qubit_id
    )
    return req1, req2


def test1():
    alice, bob = create_n_qdevices(2)
    entdist = create_entdist([alice, bob])

    request_alice = create_request(alice.node.ID, bob.node.ID)
    request_bob = create_request(bob.node.ID, alice.node.ID)

    entdist.put_request(request_alice)
    entdist.put_request(request_bob)

    ns.sim_reset()
    assert ns.sim_time() == 0
    netsquid_run(entdist.serve_all_requests())
    assert ns.sim_time() == 1000

    alice_qubit = alice.get_local_qubit(0)
    bob_qubit = bob.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)


def test2():
    qdevices = create_n_qdevices(4, num_qubits=2)
    entdist = create_entdist(qdevices)

    ids = [qdevices[i].node.ID for i in range(4)]

    req01, req10 = create_request_pair(ids[0], ids[1], 0, 0)
    entdist.put_request(req01)
    entdist.put_request(req10)

    req02, req20 = create_request_pair(ids[0], ids[2], 1, 0)
    entdist.put_request(req02)
    entdist.put_request(req20)

    req13, req31 = create_request_pair(ids[1], ids[3], 1, 0)
    entdist.put_request(req13)
    entdist.put_request(req31)

    ns.sim_reset()
    assert ns.sim_time() == 0
    netsquid_run(entdist.serve_all_requests())
    assert ns.sim_time() == 3 * 1000  # 3 request pairs

    n0_q0 = qdevices[0].get_local_qubit(0)
    n0_q1 = qdevices[0].get_local_qubit(1)
    n1_q0 = qdevices[1].get_local_qubit(0)
    n1_q1 = qdevices[1].get_local_qubit(1)
    n2_q0 = qdevices[2].get_local_qubit(0)
    n2_q1 = qdevices[2].get_local_qubit(1)
    n3_q0 = qdevices[3].get_local_qubit(0)
    n3_q1 = qdevices[3].get_local_qubit(1)

    assert has_multi_state([n0_q0, n1_q0], B00_DENS)
    assert has_multi_state([n0_q1, n2_q0], B00_DENS)
    assert has_multi_state([n1_q1, n3_q0], B00_DENS)

    assert n2_q1 is None
    assert n3_q1 is None


if __name__ == "__main__":
    test1()
    test2()



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional, Tuple

from netsquid.nodes import Node
from netsquid.protocols import Protocol
from netsquid.qubits import qubitapi
from netsquid.qubits.qrepr import QRepr
from netsquid.qubits.qubit import Qubit
from netsquid_magic.state_delivery_sampler import (
    DeliverySample,
    IStateDeliverySamplerFactory,
    StateDeliverySampler,
)

from pydynaa import EventExpression
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.lhi import LhiLinkInfo
from qoala.runtime.message import Message
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.entdist.entdistinterface import EntDistInterface
from qoala.sim.events import EPR_DELIVERY, MSG_REQUEST_DELIVERED
from qoala.util.logging import LogManager


@dataclass(eq=True, frozen=True)
class EntDistRequest:
    local_node_id: int
    remote_node_id: int
    local_qubit_id: int


@dataclass(eq=True, frozen=True)
class JointRequest:
    node1_id: int
    node2_id: int
    node1_qubit_id: int
    node2_qubit_id: int


@dataclass
class EprDeliverySample:
    state: QRepr
    duration: float

    @classmethod
    def from_ns_magic_delivery_sample(cls, sample: DeliverySample) -> EprDeliverySample:
        assert sample.state.num_qubits == 2
        return EprDeliverySample(
            state=sample.state.dm, duration=sample.delivery_duration
        )


@dataclass
class DelayedSampler:
    sampler: StateDeliverySampler
    delay: float


class EntDist(Protocol):
    def __init__(
        self, nodes: List[Node], network_info: NetworkInfo, comp: EntDistComponent
    ) -> None:
        super().__init__(name=f"{comp.name}_protocol")

        # References to objects.
        self._network_info = network_info
        self._comp = comp

        # Owned objects.
        self._interface = EntDistInterface(comp, network_info)

        # Node ID -> Node
        self._nodes: Dict[int, Node] = {node.ID: node for node in nodes}

        # (Node ID 1, Node ID 2) -> Sampler
        self._samplers: Dict[Tuple[int, int], DelayedSampler] = {}

        # Node ID -> list of requests
        self._requests: Dict[int, List[EntDistRequest]] = {
            node.ID: [] for node in nodes
        }

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}(EntDist)"
        )

    @property
    def comp(self) -> EntDistComponent:
        return self._comp

    def _add_sampler(
        self,
        node1_id: int,
        node2_id: int,
        factory: IStateDeliverySamplerFactory,
        kwargs: Dict[str, Any],
        delay: float,
    ) -> None:
        if (node1_id, node2_id) in self._samplers:
            raise ValueError(f"Sampler for ({node1_id}, {node2_id}) already registered")
        if (node2_id, node1_id) in self._samplers:
            raise ValueError(
                f"Sampler for ({node1_id}, {node2_id}) already registered. "
                "NOTE: only one sampler per node pair is allowed; order does not matter."
            )
        sampler = factory.create_state_delivery_sampler(**kwargs)
        self._samplers[(node1_id, node2_id)] = DelayedSampler(sampler, delay)

    def add_sampler(self, node1_id: int, node2_id: int, info: LhiLinkInfo) -> None:
        self._add_sampler(
            node1_id=node1_id,
            node2_id=node2_id,
            factory=info.sampler_factory(),
            kwargs=info.sampler_kwargs,
            delay=info.state_delay,
        )

    def get_sampler(self, node1_id: int, node2_id: int) -> DelayedSampler:
        try:
            return self._samplers[(node1_id, node2_id)]
        except KeyError:
            pass
        try:
            return self._samplers[(node2_id, node1_id)]
        except KeyError:
            raise ValueError(f"No sampler registered for pair ({node1_id}, {node2_id})")

    def sample_state(self, sampler: StateDeliverySampler) -> EprDeliverySample:
        raw_sample: DeliverySample = sampler.sample()
        return EprDeliverySample.from_ns_magic_delivery_sample(raw_sample)

    def create_epr_pair_with_state(self, state: QRepr) -> Tuple[Qubit, Qubit]:
        q0, q1 = qubitapi.create_qubits(2)
        qubitapi.assign_qstate([q0, q1], state)
        return q0, q1

    def deliver(
        self,
        node1_id: int,
        node1_phys_id: int,
        node2_id: int,
        node2_phys_id: int,
    ) -> Generator[EventExpression, None, None]:
        timed_sampler = self.get_sampler(node1_id, node2_id)
        sample = self.sample_state(timed_sampler.sampler)
        epr = self.create_epr_pair_with_state(sample.state)

        total_delay = sample.duration + timed_sampler.delay

        node1_mem = self._nodes[node1_id].qmemory
        node2_mem = self._nodes[node2_id].qmemory

        node1_mem.mem_positions[node1_phys_id].in_use = True
        node2_mem.mem_positions[node2_phys_id].in_use = True

        self._schedule_after(total_delay, EPR_DELIVERY)
        event_expr = EventExpression(source=self, event_type=EPR_DELIVERY)
        yield event_expr

        node1_mem.put(qubits=epr[0], positions=node1_phys_id)
        node2_mem.put(qubits=epr[1], positions=node2_phys_id)

        # Send messages to the nodes indictating a request has been delivered.
        node1 = self._interface.remote_id_to_peer_name(node1_id)
        node2 = self._interface.remote_id_to_peer_name(node2_id)
        self._interface.send_node_msg(node1, Message(MSG_REQUEST_DELIVERED))
        self._interface.send_node_msg(node2, Message(MSG_REQUEST_DELIVERED))

    def put_request(self, request: EntDistRequest) -> None:
        if request.local_node_id not in self._nodes:
            raise ValueError(
                f"Invalid request: node ID {request.local_node_id} not registered in EntDist."
            )
        if request.remote_node_id not in self._nodes:
            raise ValueError(
                f"Invalid request: node ID {request.remote_node_id} not registered in EntDist."
            )
        self._requests[request.local_node_id].append(request)

    def get_requests(self, node_id: int) -> List[EntDistRequest]:
        return self._requests[node_id]

    def pop_request(self, node_id: int, index: int) -> EntDistRequest:
        return self._requests[node_id].pop(index)

    def get_remote_request_for(self, local_request: EntDistRequest) -> Optional[int]:
        """Return index in request list of remote node."""
        remote_request_index = None

        try:
            remote_requests = self._requests[local_request.remote_node_id]
        except KeyError:
            # Invalid remote node ID.
            raise ValueError(
                f"Request {local_request} refers to remote node "
                f"{local_request.remote_node_id} "
                "but this node is not registed in the EntDist."
            )

        for i, req in enumerate(remote_requests):
            # Find the remote request that corresponds to the local request.
            if req.remote_node_id == local_request.local_node_id:
                remote_request_index = i
                break  # break out of "for i, remote_request" loop

        return remote_request_index

    def get_next_joint_request(self) -> Optional[JointRequest]:
        for _, local_requests in self._requests.items():
            if len(local_requests) == 0:
                continue
            local_request = local_requests.pop(0)
            remote_request_id = self.get_remote_request_for(local_request)
            if remote_request_id is not None:
                remote_id = local_request.remote_node_id
                remote_request = self._requests[remote_id].pop(remote_request_id)
                return JointRequest(
                    local_request.local_node_id,
                    remote_request.local_node_id,
                    local_request.local_qubit_id,
                    remote_request.local_qubit_id,
                )
            else:
                # Put local request back
                local_requests.insert(0, local_request)

        # No joint requests found
        return None

    def serve_request(
        self, request: JointRequest
    ) -> Generator[EventExpression, None, None]:
        yield from self.deliver(
            node1_id=request.node1_id,
            node1_phys_id=request.node1_qubit_id,
            node2_id=request.node2_id,
            node2_phys_id=request.node2_qubit_id,
        )

    def serve_all_requests(self) -> Generator[EventExpression, None, None]:
        while request := self.get_next_joint_request():
            yield from self.deliver(
                node1_id=request.node1_id,
                node1_phys_id=request.node1_qubit_id,
                node2_id=request.node2_id,
                node2_phys_id=request.node2_qubit_id,
            )

    def start(self) -> None:
        assert self._interface is not None
        super().start()
        self._interface.start()

    def stop(self) -> None:
        self._interface.stop()
        super().stop()

    def run(self) -> Generator[EventExpression, None, None]:
        # Loop forever acting on messages from the nodes.
        while True:
            # Wait for a new message.
            msg = yield from self._interface.receive_msg()
            self._logger.debug(f"received new msg from node: {msg}")
            request: EntDistRequest = msg.content
            self.put_request(request)
            yield from self.serve_all_requests()



from __future__ import annotations

import logging
from abc import abstractmethod
from typing import Generator, List, Optional

import netsquid as ns
from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.lang.hostlang import BasicBlockType, RunRequestOp, RunSubroutineOp
from qoala.runtime.schedule import TaskSchedule, TaskScheduleEntry
from qoala.runtime.task import BlockTask, TaskExecutionMode
from qoala.sim.events import EVENT_WAIT, SIGNAL_TASK_COMPLETED
from qoala.sim.host.hostprocessor import HostProcessor
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack.netstackprocessor import NetstackProcessor
from qoala.sim.process import QoalaProcess
from qoala.sim.qnos.qnosprocessor import QnosProcessor
from qoala.util.logging import LogManager


class Driver(Protocol):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)
        self.add_signal(SIGNAL_TASK_COMPLETED)

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({name})"
        )

        self._other_driver: Optional[Driver] = None
        self._task_list: List[TaskScheduleEntry] = []

        self._finished_tasks: List[BlockTask] = []

    def set_other_driver(self, other: Driver) -> None:
        self._other_driver = other

    def run(self) -> Generator[EventExpression, None, None]:
        while True:
            try:
                entry = self._task_list.pop(0)
                time = entry.timestamp
                task = entry.task
                prev = entry.prev
                if time is not None:
                    now = ns.sim_time()
                    self._logger.debug(
                        f"{ns.sim_time()}: {self.name}: checking next task {task}"
                    )
                    self._logger.debug(f"scheduled for {time}")
                    if time - now <= 0:
                        self._logger.debug(
                            "scheduled time is in the past, so not waiting"
                        )
                    else:
                        self._logger.debug(f"waiting for {time - now}...")
                        yield from self.wait(time - now)
                if prev is not None:
                    assert self._other_driver is not None
                    while prev not in self._other_driver._finished_tasks:
                        # Wait for a signal that the other driver completed a task.
                        yield self.await_signal(
                            sender=self._other_driver,
                            signal_label=SIGNAL_TASK_COMPLETED,
                        )

                self._logger.info(f"executing task {task}")
                yield from self._handle_task(task)
                self._finished_tasks.append(task)
                self.send_signal(SIGNAL_TASK_COMPLETED)
                self._logger.info(f"finished task {task}")
            except IndexError:
                break

    @abstractmethod
    def _handle_task(self, task: BlockTask) -> Generator[EventExpression, None, None]:
        raise NotImplementedError


class CpuDriver(Driver):
    def __init__(
        self,
        node_name: str,
        hostprocessor: HostProcessor,
        memmgr: MemoryManager,
    ) -> None:
        super().__init__(name=f"{node_name}_cpu_driver")

        self._hostprocessor = hostprocessor
        self._memmgr = memmgr

    def upload_schedule(self, schedule: TaskSchedule) -> None:
        self._task_list.extend(schedule.entries)

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr

    def _handle_task(self, task: BlockTask) -> Generator[EventExpression, None, None]:
        process = self._memmgr.get_process(task.pid)
        yield from self._hostprocessor.assign_block(process, task.block_name)


class QpuDriver(Driver):
    def __init__(
        self,
        node_name: str,
        hostprocessor: HostProcessor,
        qnosprocessor: QnosProcessor,
        netstackprocessor: NetstackProcessor,
        memmgr: MemoryManager,
        tem: TaskExecutionMode = TaskExecutionMode.ROUTINE_ATOMIC,
    ) -> None:
        super().__init__(name=f"{node_name}_qpu_driver")

        self._hostprocessor = hostprocessor
        self._qnosprocessor = qnosprocessor
        self._netstackprocessor = netstackprocessor
        self._memmgr = memmgr
        self._tem = tem

    def upload_schedule(self, schedule: TaskSchedule) -> None:
        self._task_list.extend(schedule.entries)

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr

    def _handle_lr(self, task: BlockTask) -> Generator[EventExpression, None, None]:
        if self._tem == TaskExecutionMode.ROUTINE_ATOMIC:
            yield from self._handle_atomic_lr(task)
        else:
            raise NotImplementedError

    def _handle_rr(self, task: BlockTask) -> Generator[EventExpression, None, None]:
        if self._tem == TaskExecutionMode.ROUTINE_ATOMIC:
            yield from self._handle_atomic_rr(task)
        else:
            raise NotImplementedError

    def allocate_qubits_for_routine(
        self, process: QoalaProcess, routine_name: str
    ) -> None:
        # TODO: merge with code in scheduler.py?
        routine = process.get_local_routine(routine_name)
        for virt_id in routine.metadata.qubit_use:
            if self._memmgr.phys_id_for(process.pid, virt_id) is None:
                self._memmgr.allocate(process.pid, virt_id)

    def free_qubits_after_routine(
        self, process: QoalaProcess, routine_name: str
    ) -> None:
        # TODO: merge with code in scheduler.py?
        routine = process.get_local_routine(routine_name)
        for virt_id in routine.metadata.qubit_use:
            if virt_id not in routine.metadata.qubit_keep:
                self._memmgr.free(process.pid, virt_id)

    def _handle_atomic_lr(
        self, task: BlockTask
    ) -> Generator[EventExpression, None, None]:
        process = self._memmgr.get_process(task.pid)
        block = process.program.get_block(task.block_name)
        assert len(block.instructions) == 1
        instr = block.instructions[0]
        assert isinstance(instr, RunSubroutineOp)

        # Let Host setup shared memory.
        lrcall = self._hostprocessor.prepare_lr_call(process, instr)
        # Allocate required qubits.
        self.allocate_qubits_for_routine(process, lrcall.routine_name)
        # Execute the routine on Qnos.
        yield from self._qnosprocessor.assign_local_routine(
            process, lrcall.routine_name, lrcall.input_addr, lrcall.result_addr
        )
        # Free qubits that do not need to be kept.
        self.free_qubits_after_routine(process, lrcall.routine_name)
        # Let Host get results from shared memory.
        self._hostprocessor.post_lr_call(process, instr, lrcall)

    def _handle_atomic_rr(
        self, task: BlockTask
    ) -> Generator[EventExpression, None, None]:
        process = self._memmgr.get_process(task.pid)
        block = process.program.get_block(task.block_name)
        assert len(block.instructions) == 1
        instr = block.instructions[0]
        assert isinstance(instr, RunRequestOp)

        # Let Host setup shared memory.
        rrcall = self._hostprocessor.prepare_rr_call(process, instr)
        # TODO: refactor this. Bit of a hack to just pass the QnosProcessor around like this!
        yield from self._netstackprocessor.assign_request_routine(
            process, rrcall, self._qnosprocessor
        )
        self._hostprocessor.post_rr_call(process, instr, rrcall)

    def _handle_task(self, task: BlockTask) -> Generator[EventExpression, None, None]:
        if task.typ == BasicBlockType.QL:
            yield from self._handle_lr(task)
        elif task.typ == BasicBlockType.QC:
            yield from self._handle_rr(task)
        else:
            raise RuntimeError



import logging
from dataclasses import dataclass
from typing import Dict, Optional

from qoala.lang.ehi import EhiNodeInfo, UnitModule
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice
from qoala.util.logging import LogManager


class AllocError(Exception):
    """Allocating a qubit did not succeed."""

    pass


class NotAllocatedError(Exception):
    """A virtual qubit is not mapped to any physical qubit."""

    pass


@dataclass
class VirtualMapping:
    # mapping from virt ID in specific unit module to phys ID
    unit_module: UnitModule
    mapping: Dict[int, Optional[int]]  # virt ID -> phys ID


@dataclass
class VirtualLocation:
    # particular virt ID in particular unit module in particular process
    pid: int
    unit_module: UnitModule
    virt_id: int


class MemoryManager:
    def __init__(
        self,
        node_name: str,
        qdevice: QDevice,
        ehi: Optional[EhiNodeInfo] = None,  # TODO refactor?
    ) -> None:
        self._node_name = node_name
        self._processes: Dict[int, QoalaProcess] = {}
        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._node_name})"
        )
        self._ehi = ehi

        self._qdevice = qdevice
        self._process_mappings: Dict[int, VirtualMapping] = {}  # pid -> mapping
        self._physical_mapping: Dict[int, Optional[VirtualLocation]] = {
            i: None for i in qdevice.get_all_qubit_ids()
        }  # phys ID -> virt location

    def _get_free_comm_phys_id(self) -> int:
        for phys_id in self._qdevice.get_comm_qubit_ids():
            if self._physical_mapping[phys_id] is None:
                return phys_id  # type: ignore
        raise AllocError

    def _get_free_mem_phys_id(self) -> int:
        for phys_id in self._qdevice.get_non_comm_qubit_ids():
            if self._physical_mapping[phys_id] is None:
                return phys_id  # type: ignore
        raise AllocError

    def get_ehi(self) -> EhiNodeInfo:
        assert self._ehi is not None  # TODO: already enforce this in constructor?
        return self._ehi

    def add_process(self, process: QoalaProcess) -> None:
        self._processes[process.pid] = process
        unit_module = process.prog_instance.unit_module
        self._process_mappings[process.pid] = VirtualMapping(
            unit_module, {x: None for x in unit_module.get_all_qubit_ids()}
        )

    def get_process(self, pid: int) -> QoalaProcess:
        return self._processes[pid]

    def allocate(self, pid: int, virt_id: int) -> int:
        vmap = self._process_mappings[pid]
        # Check if the virtual ID is in the unit module
        if virt_id not in vmap.unit_module.get_all_qubit_ids():
            raise AllocError

        # Check whether this virt ID is already mapped to a physical qubit.
        if vmap.mapping[virt_id] is not None:
            raise AllocError

        phys_id: int
        if vmap.unit_module.is_communication(virt_id):
            phys_id = self._get_free_comm_phys_id()
        else:
            phys_id = self._get_free_mem_phys_id()

        # update mappings
        self._physical_mapping[phys_id] = VirtualLocation(
            pid, vmap.unit_module, virt_id
        )
        self._process_mappings[pid].mapping[virt_id] = phys_id
        return phys_id

    def allocate_comm(self, pid: int, virt_id: int) -> int:
        vmap = self._process_mappings[pid]
        # Check that the virt ID is indeed a (virtual) comm qubit.
        if virt_id not in vmap.unit_module.get_all_qubit_ids():
            raise AllocError
        if not vmap.unit_module.is_communication(virt_id):
            raise AllocError

        return self.allocate(pid, virt_id)

    def free(self, pid: int, virt_id: int) -> None:
        vmap = self._process_mappings[pid]
        # Check if the virtual ID is in the unit module
        assert virt_id in vmap.unit_module.get_all_qubit_ids()
        assert virt_id in self._process_mappings[pid].mapping

        phys_id = self._process_mappings[pid].mapping[virt_id]
        if phys_id is None:
            raise AllocError

        # update mappings
        self._physical_mapping[phys_id] = None
        self._process_mappings[pid].mapping[virt_id] = None

        # update netsquid memory
        self._qdevice.set_mem_pos_in_use(phys_id, False)

    def get_unmapped_non_comm_qubit(self, pid: int) -> int:
        """returns virt ID"""
        vp_map = self._process_mappings[pid].mapping
        unit_module = self._process_mappings[pid].unit_module
        free_ids = [
            v
            for v, p in vp_map.items()
            if p is None and not unit_module.is_communication(v)
        ]
        if len(free_ids) == 0:
            raise AllocError
        return min(free_ids)

    def phys_id_for(self, pid: int, virt_id: int) -> Optional[int]:
        phys_id = self._process_mappings[pid].mapping[virt_id]
        return phys_id

    def virt_id_for(self, pid: int, phys_id: int) -> Optional[int]:
        if virt_loc := self._physical_mapping[phys_id]:
            if virt_loc.pid == pid:
                return virt_loc.virt_id
        return None



from __future__ import annotations

from netsquid.protocols import Protocol

from qoala.runtime.environment import LocalEnvironment
from qoala.runtime.lhi_to_ehi import (
    GenericToVanillaInterface,
    NativeToFlavourInterface,
    NvToNvInterface,
)
from qoala.sim.memmgr import MemoryManager
from qoala.sim.qdevice import QDevice
from qoala.sim.qnos.qnoscomp import QnosComponent
from qoala.sim.qnos.qnosinterface import QnosInterface, QnosLatencies
from qoala.sim.qnos.qnosprocessor import GenericProcessor, NVProcessor, QnosProcessor


class Qnos(Protocol):
    """NetSquid protocol representing a QNodeOS instance."""

    def __init__(
        self,
        comp: QnosComponent,
        local_env: LocalEnvironment,
        memmgr: MemoryManager,
        qdevice: QDevice,
        latencies: QnosLatencies,
        ntf_interface: NativeToFlavourInterface,
        asynchronous: bool = False,
    ) -> None:
        """Qnos protocol constructor.

        :param comp: NetSquid component representing the QNodeOS instance
        :param qdevice_type: hardware type of the QDevice of this node
        """
        super().__init__(name=f"{comp.name}_protocol")

        # References to objects.
        self._comp = comp
        self._local_env = local_env

        # Owned objects.
        self._interface = QnosInterface(comp, qdevice, memmgr)
        self._processor: QnosProcessor
        self._asynchronous = asynchronous

        self.create_processor(ntf_interface, latencies)

    def create_processor(
        self, ntf_interface: NativeToFlavourInterface, latencies: QnosLatencies
    ) -> None:
        # TODO: rethink the way NTF interfaces are used
        if isinstance(ntf_interface, GenericToVanillaInterface):
            self._processor = GenericProcessor(
                self._interface, latencies, self._asynchronous
            )
        elif isinstance(ntf_interface, NvToNvInterface):
            self._processor = NVProcessor(
                self._interface, latencies, self._asynchronous
            )
        else:
            raise ValueError

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    @qdevice.setter
    def qdevice(self, qdevice: QDevice) -> None:
        self._interface._qdevice = qdevice

    @property
    def processor(self) -> QnosProcessor:
        return self._processor

    @processor.setter
    def processor(self, processor: QnosProcessor) -> None:
        self._processor = processor

    def start(self) -> None:
        super().start()
        self._interface.start()

    def stop(self) -> None:
        self._interface.stop()
        super().stop()



import netsquid as ns
import pytest

from qoala.lang.hostlang import BasicBlockType
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.schedule import BlockTask, TaskSchedule, TaskScheduleEntry
from qoala.sim.driver import CpuDriver, QpuDriver
from qoala.util.builder import ObjectBuilder


def get_pure_host_program() -> QoalaProgram:
    program_text = """
META_START
    name: alice
    parameters:
    csockets:
    epr_sockets:
META_END

^b0 {type = CL}:
    var_x = assign_cval() : 3
    var_y = assign_cval() : 5
^b1 {type = CL}:
    var_z = assign_cval() : 9
    """

    return QoalaParser(program_text).parse()


def get_lr_program() -> QoalaProgram:
    program_text = """
META_START
    name: alice
    parameters:
    csockets:
    epr_sockets:
META_END

^b0 {type = CL}:
    x = assign_cval() : 3
^b1 {type = QL}:
    vec<y> = run_subroutine(vec<x>) : add_one

SUBROUTINE add_one
    params: x
    returns: y
    uses: 
    keeps:
    request:
  NETQASM_START
    load C0 @input[0]
    set C1 1
    add R0 C0 C1
    store R0 @output[0]
  NETQASM_END
    """

    return QoalaParser(program_text).parse()


CL = BasicBlockType.CL
QL = BasicBlockType.QL
QC = BasicBlockType.QC


def test_cpu_driver():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_pure_host_program()

    pid = 0
    instance = ObjectBuilder.simple_program_instance(program, pid)

    procnode.scheduler.submit_program_instance(instance)

    cpu_schedule = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(0, "b0", CL), 0),
            TaskScheduleEntry(BlockTask(0, "b1", CL), 1000),
        ]
    )

    driver = CpuDriver("alice", procnode.host.processor, procnode.memmgr)
    driver.upload_schedule(cpu_schedule)

    ns.sim_reset()
    driver.start()
    ns.sim_run()

    assert procnode.memmgr.get_process(pid).host_mem.read("var_x") == 3
    assert procnode.memmgr.get_process(pid).host_mem.read("var_y") == 5
    assert procnode.memmgr.get_process(pid).host_mem.read("var_z") == 9

    assert ns.sim_time() == 1000


def test_cpu_driver_no_time():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_pure_host_program()

    pid = 0
    instance = ObjectBuilder.simple_program_instance(program, pid)

    procnode.scheduler.submit_program_instance(instance)

    cpu_schedule = TaskSchedule.consecutive(
        [BlockTask(0, "b0", CL), BlockTask(0, "b1", CL)]
    )

    driver = CpuDriver("alice", procnode.host.processor, procnode.memmgr)
    driver.upload_schedule(cpu_schedule)

    ns.sim_reset()
    driver.start()
    ns.sim_run()

    assert procnode.memmgr.get_process(pid).host_mem.read("var_x") == 3
    assert procnode.memmgr.get_process(pid).host_mem.read("var_y") == 5
    assert procnode.memmgr.get_process(pid).host_mem.read("var_z") == 9

    assert ns.sim_time() == 0


def test_cpu_driver_2_processes():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_pure_host_program()

    pid0 = 0
    pid1 = 1
    instance0 = ObjectBuilder.simple_program_instance(program, pid0)
    instance1 = ObjectBuilder.simple_program_instance(program, pid1)

    procnode.scheduler.submit_program_instance(instance0)
    procnode.scheduler.submit_program_instance(instance1)

    cpu_schedule = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(pid0, "b0", CL), 0),
            TaskScheduleEntry(BlockTask(pid1, "b0", CL), 500),
            TaskScheduleEntry(BlockTask(pid0, "b1", CL), 1000),
            TaskScheduleEntry(BlockTask(pid1, "b1", CL), 1500),
        ]
    )

    driver = CpuDriver("alice", procnode.host.processor, procnode.memmgr)
    driver.upload_schedule(cpu_schedule)

    ns.sim_reset()
    driver.start()
    ns.sim_run(duration=1000)

    assert procnode.memmgr.get_process(pid0).host_mem.read("var_x") == 3
    assert procnode.memmgr.get_process(pid0).host_mem.read("var_y") == 5
    with pytest.raises(KeyError):
        procnode.memmgr.get_process(pid0).host_mem.read("var_z")
        procnode.memmgr.get_process(pid1).host_mem.read("var_z")

    ns.sim_run()
    assert procnode.memmgr.get_process(pid0).host_mem.read("var_z") == 9
    assert procnode.memmgr.get_process(pid1).host_mem.read("var_x") == 3
    assert procnode.memmgr.get_process(pid1).host_mem.read("var_y") == 5
    assert procnode.memmgr.get_process(pid1).host_mem.read("var_z") == 9

    assert ns.sim_time() == 1500


def test_qpu_driver():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_lr_program()

    pid = 0
    instance = ObjectBuilder.simple_program_instance(program, pid)

    procnode.scheduler.submit_program_instance(instance)

    cpu_schedule = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(0, "b0", CL), 0),
        ]
    )
    qpu_schedule = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(0, "b1", QL), 1000),
        ]
    )

    cpudriver = CpuDriver("alice", procnode.host.processor, procnode.memmgr)
    cpudriver.upload_schedule(cpu_schedule)

    qpudriver = QpuDriver(
        "alice",
        procnode.host.processor,
        procnode.qnos.processor,
        procnode.netstack.processor,
        procnode.memmgr,
    )
    qpudriver.upload_schedule(qpu_schedule)

    ns.sim_reset()
    cpudriver.start()
    qpudriver.start()
    ns.sim_run()

    assert procnode.memmgr.get_process(pid).host_mem.read("y") == 4

    assert ns.sim_time() == 1000


def test_qpu_driver_2_processes():
    procnode = ObjectBuilder.simple_procnode("alice", 1)
    program = get_lr_program()

    pid0 = 0
    pid1 = 1
    instance0 = ObjectBuilder.simple_program_instance(program, pid0)
    instance1 = ObjectBuilder.simple_program_instance(program, pid1)

    procnode.scheduler.submit_program_instance(instance0)
    procnode.scheduler.submit_program_instance(instance1)

    cpu_schedule = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(pid0, "b0", CL), 0),
            TaskScheduleEntry(BlockTask(pid1, "b0", CL), 500),
        ]
    )
    qpu_schedule = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(pid0, "b1", QL), 500),
            TaskScheduleEntry(BlockTask(pid1, "b1", QL), 1000),
        ]
    )

    cpudriver = CpuDriver("alice", procnode.host.processor, procnode.memmgr)
    cpudriver.upload_schedule(cpu_schedule)

    qpudriver = QpuDriver(
        "alice",
        procnode.host.processor,
        procnode.qnos.processor,
        procnode.netstack.processor,
        procnode.memmgr,
    )
    qpudriver.upload_schedule(qpu_schedule)

    ns.sim_reset()
    cpudriver.start()
    qpudriver.start()
    ns.sim_run()

    assert procnode.memmgr.get_process(pid0).host_mem.read("y") == 4
    assert procnode.memmgr.get_process(pid1).host_mem.read("y") == 4

    assert ns.sim_time() == 1000


if __name__ == "__main__":
    test_cpu_driver()
    test_cpu_driver_no_time()
    test_cpu_driver_2_processes()
    test_qpu_driver()
    test_qpu_driver_2_processes()



from __future__ import annotations

from typing import Generator, List

import netsquid as ns
from netsquid.nodes import Node

from pydynaa import EventExpression
from qoala.runtime.environment import LocalEnvironment, NetworkInfo
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.entdist.entdistinterface import EntDistInterface
from qoala.sim.netstack.netstackcomp import NetstackComponent
from qoala.sim.netstack.netstackinterface import NetstackInterface


class MockNetstackInterface(NetstackInterface):
    def __init__(self, comp: NetstackComponent, local_env: LocalEnvironment) -> None:
        super().__init__(comp, local_env, None, None)


def test_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    env = NetworkInfo.with_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = NetstackComponent(alice, env)
    bob_comp = NetstackComponent(bob, env)
    entdist_comp = EntDistComponent(env)

    # Connect both nodes to the Entdist.
    alice_comp.entdist_out_port.connect(entdist_comp.node_in_port("alice"))
    alice_comp.entdist_in_port.connect(entdist_comp.node_out_port("alice"))
    bob_comp.entdist_out_port.connect(entdist_comp.node_in_port("bob"))
    bob_comp.entdist_in_port.connect(entdist_comp.node_out_port("bob"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_entdist_msg("hello this is Alice")

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_entdist_msg("hello this is Bob")

    class TestEntDistInterface(EntDistInterface):
        def __init__(self, comp: EntDistComponent, env: NetworkInfo) -> None:
            super().__init__(comp, env)
            self.msg_alice = None
            self.msg_bob = None

        def run(self) -> Generator[EventExpression, None, None]:
            self.msg_alice = yield from self.receive_node_msg("alice")
            self.msg_bob = yield from self.receive_node_msg("bob")

    alice_intf = AliceNetstackInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobNetstackInterface(bob_comp, LocalEnvironment(env, bob.ID))
    entdist_intf = TestEntDistInterface(entdist_comp, env)

    alice_intf.start()
    bob_intf.start()
    entdist_intf.start()

    ns.sim_run()

    assert entdist_intf.msg_alice == "hello this is Alice"
    assert entdist_intf.msg_bob == "hello this is Bob"


def test_wait_for_any_node():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    env = NetworkInfo.with_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = NetstackComponent(alice, env)
    bob_comp = NetstackComponent(bob, env)
    entdist_comp = EntDistComponent(env)

    # Connect both nodes to the Entdist.
    alice_comp.entdist_out_port.connect(entdist_comp.node_in_port("alice"))
    alice_comp.entdist_in_port.connect(entdist_comp.node_out_port("alice"))
    bob_comp.entdist_out_port.connect(entdist_comp.node_in_port("bob"))
    bob_comp.entdist_in_port.connect(entdist_comp.node_out_port("bob"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_entdist_msg("hello this is Alice")
            yield from self.wait(2000)
            self.send_entdist_msg("hello again from Alice")

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_entdist_msg("hello this is Bob")
            yield from self.wait(1000)
            self.send_entdist_msg("hello again from Bob")

    class TestEntDistInterface(EntDistInterface):
        def __init__(self, comp: EntDistComponent, env: NetworkInfo) -> None:
            super().__init__(comp, env)
            self.messages: List[str] = []

        def run(self) -> Generator[EventExpression, None, None]:
            self.messages.append((yield from self.receive_msg()))
            self.messages.append((yield from self.receive_msg()))
            self.messages.append((yield from self.receive_msg()))
            self.messages.append((yield from self.receive_msg()))

    alice_intf = AliceNetstackInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobNetstackInterface(bob_comp, LocalEnvironment(env, bob.ID))
    entdist_intf = TestEntDistInterface(entdist_comp, env)

    alice_intf.start()
    bob_intf.start()
    entdist_intf.start()

    ns.sim_run()

    assert entdist_intf.messages == [
        "hello this is Alice",
        "hello this is Bob",
        "hello again from Bob",
        "hello again from Alice",
    ]


def test_wait_for_any_node_2():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    charlie = Node(name="charlie", ID=2)
    env = NetworkInfo.with_nodes(
        {alice.ID: alice.name, bob.ID: bob.name, charlie.ID: charlie.name}
    )

    alice_comp = NetstackComponent(alice, env)
    bob_comp = NetstackComponent(bob, env)
    charlie_comp = NetstackComponent(charlie, env)
    entdist_comp = EntDistComponent(env)

    # Connect all nodes to the Entdist.
    alice_comp.entdist_out_port.connect(entdist_comp.node_in_port("alice"))
    alice_comp.entdist_in_port.connect(entdist_comp.node_out_port("alice"))
    bob_comp.entdist_out_port.connect(entdist_comp.node_in_port("bob"))
    bob_comp.entdist_in_port.connect(entdist_comp.node_out_port("bob"))
    charlie_comp.entdist_out_port.connect(entdist_comp.node_in_port("charlie"))
    charlie_comp.entdist_in_port.connect(entdist_comp.node_out_port("charlie"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            yield from self.wait(500)
            self.send_entdist_msg("hello this is Alice at time 500")
            yield from self.wait(2000)
            self.send_entdist_msg("hello again from Alice at time 2500")

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_entdist_msg("hello this is Bob at time 0")
            yield from self.wait(1000)
            self.send_entdist_msg("hello again from Bob at time 1000")

    class CharlieNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            yield from self.wait(300)
            self.send_entdist_msg("hello this is Charlie at time 300")
            yield from self.wait(1500)
            self.send_entdist_msg("hello again from Charlie at time 1800")

    class TestEntDistInterface(EntDistInterface):
        def __init__(self, comp: EntDistComponent, env: NetworkInfo) -> None:
            super().__init__(comp, env)
            self.messages: List[str] = []

        def run(self) -> Generator[EventExpression, None, None]:
            for _ in range(6):
                self.messages.append((yield from self.receive_msg()))

    alice_intf = AliceNetstackInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobNetstackInterface(bob_comp, LocalEnvironment(env, bob.ID))
    charlie_intf = CharlieNetstackInterface(
        charlie_comp, LocalEnvironment(env, charlie.ID)
    )
    entdist_intf = TestEntDistInterface(entdist_comp, env)

    alice_intf.start()
    bob_intf.start()
    charlie_intf.start()
    entdist_intf.start()

    ns.sim_run()

    assert entdist_intf.messages == [
        "hello this is Bob at time 0",
        "hello this is Charlie at time 300",
        "hello this is Alice at time 500",
        "hello again from Bob at time 1000",
        "hello again from Charlie at time 1800",
        "hello again from Alice at time 2500",
    ]


if __name__ == "__main__":
    test_connection()
    test_wait_for_any_node()
    test_wait_for_any_node_2()



from __future__ import annotations

from typing import Dict, List, Optional, Tuple

import netsquid as ns
from netqasm.lang.parsing import parse_text_subroutine
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_H,
    INSTR_I,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.models.qerrormodels import DepolarNoiseModel, T1T2NoiseModel
from netsquid.components.qprocessor import PhysicalInstruction, QuantumProcessor
from netsquid.components.qprogram import QuantumProgram
from netsquid.nodes import Node
from netsquid.qubits import ketstates, qubitapi

from qoala.lang.ehi import UnitModule
from qoala.lang.program import LocalRoutine, ProgramMeta, QoalaProgram
from qoala.lang.routine import RoutineMetadata
from qoala.runtime.lhi import LhiTopologyBuilder
from qoala.runtime.lhi_to_ehi import GenericToVanillaInterface, LhiConverter
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.sharedmem import MemAddr
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.memmgr import MemoryManager
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice
from qoala.sim.qnos import (
    GenericProcessor,
    QnosComponent,
    QnosInterface,
    QnosLatencies,
    QnosProcessor,
)
from qoala.util.math import has_max_mixed_state, has_state
from qoala.util.tests import netsquid_run, netsquid_wait


def uniform_qdevice_noisy_qubits(num_qubits: int) -> QDevice:
    qubit_info = LhiTopologyBuilder.t1t2_qubit(is_communication=True, t1=1, t2=1)
    single_gate_infos = LhiTopologyBuilder.perfect_gates(
        duration=1e4,
        instructions=[
            INSTR_INIT,
            INSTR_X,
            INSTR_Y,
            INSTR_Z,
            INSTR_H,
            INSTR_ROT_X,
            INSTR_ROT_Y,
            INSTR_ROT_Z,
            INSTR_MEASURE,
        ],
    )
    two_gate_infos = LhiTopologyBuilder.perfect_gates(
        duration=1e5,
        instructions=[
            INSTR_CNOT,
        ],
    )

    topology = LhiTopologyBuilder.fully_uniform(
        num_qubits=num_qubits,
        qubit_info=qubit_info,
        single_gate_infos=single_gate_infos,
        two_gate_infos=two_gate_infos,
    )
    processor = build_qprocessor_from_topology(name="processor", topology=topology)
    node = Node(name="alice", qmemory=processor)
    return QDevice(node=node, topology=topology)


def setup_noisy_components(num_qubits: int) -> Tuple[QnosProcessor, UnitModule]:
    qdevice = uniform_qdevice_noisy_qubits(num_qubits)
    ehi = LhiConverter.to_ehi(qdevice.topology, ntf=GenericToVanillaInterface())
    unit_module = UnitModule.from_full_ehi(ehi)
    qnos_comp = QnosComponent(node=qdevice._node)
    memmgr = MemoryManager(qdevice._node.name, qdevice)
    interface = QnosInterface(qnos_comp, qdevice, memmgr)
    processor = GenericProcessor(interface, latencies=QnosLatencies.all_zero())
    return (processor, unit_module)


def create_program(
    subroutines: Optional[Dict[str, LocalRoutine]] = None,
    meta: Optional[ProgramMeta] = None,
) -> QoalaProgram:
    if subroutines is None:
        subroutines = {}
    if meta is None:
        meta = ProgramMeta.empty("prog")
    return QoalaProgram(blocks=[], local_routines=subroutines, meta=meta)


def create_process(
    pid: int, program: QoalaProgram, unit_module: UnitModule
) -> QoalaProcess:
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=ProgramInput({}),
        unit_module=unit_module,
        block_tasks=[],
    )
    mem = ProgramMemory(pid=pid)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def create_process_with_subrt(
    pid: int,
    subrt_text: str,
    unit_module: UnitModule,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> QoalaProcess:
    if uses is None:
        uses = []
    if keeps is None:
        keeps = []
    subrt = parse_text_subroutine(subrt_text)
    metadata = RoutineMetadata(qubit_use=uses, qubit_keep=keeps)
    iqoala_subrt = LocalRoutine("subrt", subrt, return_vars=[], metadata=metadata)
    meta = ProgramMeta.empty("alice")
    meta.epr_sockets = {0: "bob"}
    program = create_program(subroutines={"subrt": iqoala_subrt}, meta=meta)
    return create_process(pid, program, unit_module)


def set_new_subroutine(process: QoalaProcess, subrt_text: str) -> None:
    subrt = parse_text_subroutine(subrt_text)
    metadata = RoutineMetadata.use_none()
    iqoala_subrt = LocalRoutine("subrt", subrt, return_vars=[], metadata=metadata)
    program = process.prog_instance.program
    program.local_routines["subrt"] = iqoala_subrt


def test_depolarizing_decoherence():
    ns.sim_reset()
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    q = qubitapi.create_qubits(1)[0]
    qubitapi.assign_qstate(q, ketstates.s1)
    print(ns.sim_time())
    assert has_state(q, ketstates.s1)
    ns.qubits.delay_depolarize(q, 1e6, delay=1e9)
    print(ns.sim_time())
    assert has_max_mixed_state(q)


def test_depolarizing_decoherence_qprocessor():
    ns.sim_reset()
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    phys_instructions = [
        PhysicalInstruction(INSTR_INIT, duration=1e3),
        PhysicalInstruction(INSTR_X, duration=1e3),
        PhysicalInstruction(INSTR_I, duration=1e10),
        PhysicalInstruction(INSTR_Z, duration=1),
    ]
    mem_noise_models = [DepolarNoiseModel(1e10)]
    processor = QuantumProcessor(
        "processor",
        num_positions=1,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )
    assert ns.sim_time() == 0
    processor.execute_instruction(INSTR_INIT, [0])
    ns.sim_run()
    assert ns.sim_time() == 1e3
    q = processor.peek([0])[0]
    assert has_state(q, ketstates.s0)

    processor.execute_instruction(INSTR_X, [0])
    ns.sim_run()
    assert ns.sim_time() == 2e3
    q = processor.peek([0])[0]
    assert has_state(q, ketstates.s1)

    processor.execute_instruction(INSTR_I, [0])
    ns.sim_run()
    assert ns.sim_time() == 1e10 + 2e3
    q = processor.peek([0])[0]
    # The I instruction made the simulator jump a whole 1e10 ns forward.
    # However, NetSquid does not apply any decoherence noise since any noise
    # is applied by the gate execution itself (in this case: no noise).
    # So, still state |1> is expected.
    assert has_state(q, ketstates.s1)

    netsquid_wait(1e10)
    assert ns.sim_time() == 2e10 + 2e3
    q = processor.peek([0])[0]
    # Now we waited another 1e10 ns, but since it was not because of executing
    # an instruction, decoherence *is* applied.
    assert has_max_mixed_state(q)


def test_depolarizing_decoherence_qprocessor_2():
    ns.sim_reset()
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    phys_instructions = [
        PhysicalInstruction(INSTR_INIT, duration=1e3),
        PhysicalInstruction(INSTR_X, duration=1e3),
        PhysicalInstruction(INSTR_I, duration=1e10),
        PhysicalInstruction(INSTR_Z, duration=1),
    ]
    mem_noise_models = [DepolarNoiseModel(1e10)]
    processor = QuantumProcessor(
        "processor",
        num_positions=1,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )
    prog = QuantumProgram()
    prog.apply(INSTR_INIT, [0])
    prog.apply(INSTR_X, [0])
    processor.execute_program(prog)
    ns.sim_run()

    # Wait 1e10 ns and then do an I instruction which takes another 1e10 ns.
    netsquid_wait(1e10)

    # NOTE: executing the following code would result in *not* applying any
    # decoherence noise, and the state would still be in |1> !
    # This is because decoherence noise is only applied when accessing the qubit
    # (like `peek` or executing an instruction), and the noise itself depends on
    # the difference between current time and 'last access time' of the qubit.
    # A `peek` with skip_noise=True will hence update the 'last access time' without
    # applying noise, and therefore the following I instruction won't apply any noise
    # either (since the 'last access time' is the same as the current time).
    # Without the following statement (i.e. as it is commented out like now),
    # the I instruction would first result in applying decoherence noise
    # (since last access time is not yet updated). Then, potentially some noise
    # is applied because of the I instruction itself (in this example: no noise).
    #
    # processor.peek([0], skip_noise=True)[0]

    prog = QuantumProgram()
    prog.apply(INSTR_I, [0])
    processor.execute_program(prog)
    ns.sim_run()
    print(ns.sim_time())
    assert ns.sim_time() == 2e10 + 2e3

    # Just before executing the I instruction, the decoherence noise (over 1e10 ns)
    # was applied, since the 'last access time' of the qubit was 1e10 ns ago.
    q = processor.peek([0])[0]
    assert has_max_mixed_state(q)


def test_t1t2_decoherence_qprocessor():
    ns.sim_reset()
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    phys_instructions = [
        PhysicalInstruction(INSTR_INIT, duration=1e3),
        PhysicalInstruction(INSTR_X, duration=1e3),
        PhysicalInstruction(INSTR_I, duration=1e10),
        PhysicalInstruction(INSTR_Z, duration=1),
    ]
    mem_noise_models = [T1T2NoiseModel(T1=10e6, T2=1e6)]
    processor = QuantumProcessor(
        "processor",
        num_positions=1,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )
    assert ns.sim_time() == 0
    processor.execute_instruction(INSTR_INIT, [0])
    ns.sim_run()
    assert ns.sim_time() == 1e3
    q = processor.peek([0])[0]
    assert has_state(q, ketstates.s0)

    processor.execute_instruction(INSTR_X, [0])
    ns.sim_run()
    assert ns.sim_time() == 2e3
    q = processor.peek([0])[0]
    assert has_state(q, ketstates.s1)

    processor.execute_instruction(INSTR_I, [0])
    ns.sim_run()
    assert ns.sim_time() == 1e10 + 2e3
    q = processor.peek([0])[0]
    # The I instruction made the simulator jump a whole 1e10 ns forward.
    # However, NetSquid does not apply any decoherence noise since any noise
    # is applied by the gate execution itself (in this case: no noise).
    # So, still state |1> is expected.
    assert has_state(q, ketstates.s1)

    netsquid_wait(1e10)
    assert ns.sim_time() == 2e10 + 2e3
    q = processor.peek([0])[0]
    # Now we waited another 1e10 ns, but since it was not because of executing
    # an instruction, decoherence *is* applied.
    assert has_state(q, ketstates.s0)


def test_decoherence_in_subroutine():
    ns.sim_reset()
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    num_qubits = 3
    processor, unit_module = setup_noisy_components(num_qubits)

    subrt = """
    set Q0 0
    init Q0
    x Q0
    """

    process = create_process_with_subrt(0, subrt, unit_module, uses=[0], keeps=[0])
    processor._interface.memmgr.add_process(process)
    processor._interface.memmgr.allocate(process.pid, 0)
    netsquid_run(
        processor.assign_local_routine(process, "subrt", MemAddr(0), MemAddr(1))
    )

    # Check if qubit with virt ID 0 has been initialized.
    phys_id = processor._interface.memmgr.phys_id_for(process.pid, virt_id=0)

    # Qubit should be in |1>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s1)

    netsquid_wait(1e9)
    qubit = processor.qdevice.get_local_qubit(phys_id)
    # Qubit should be fully dephased and amplitude-dampened.
    assert has_state(qubit, ketstates.s0)


if __name__ == "__main__":
    test_depolarizing_decoherence()
    test_depolarizing_decoherence_qprocessor()
    test_depolarizing_decoherence_qprocessor_2()
    test_t1t2_decoherence_qprocessor()
    test_decoherence_in_subroutine()
